"""
:codeauthor: Shane Lee <slee@saltstack.com>
"""

import glob
import logging
import os
import pathlib
import re
import shutil
import zipfile

import pytest
import requests

import salt.grains.core
import salt.modules.win_file as win_file
import salt.modules.win_lgpo as win_lgpo
import salt.utils.files
import salt.utils.versions
import salt.utils.win_dacl as win_dacl
from tests.support.mock import patch

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
    pytest.mark.slow_test,
]


@pytest.fixture
def configure_loader_modules(tmp_path):
    cachedir = tmp_path / "__test_admx_policy_cache_dir"
    cachedir.mkdir(parents=True, exist_ok=True)
    return {
        win_lgpo: {
            "__salt__": {
                "file.file_exists": win_file.file_exists,
                "file.makedirs": win_file.makedirs_,
            },
            "__opts__": {
                "cachedir": str(cachedir),
            },
        },
        win_file: {
            "__utils__": {
                "dacl.set_perms": win_dacl.set_perms,
            },
        },
    }


@pytest.fixture(scope="module")
def osversion():
    grains = salt.grains.core.os_data()
    yield grains.get("osversion", None)


@pytest.fixture
def clean_comp():
    reg_pol = pathlib.Path(
        os.getenv("SystemRoot"), "System32", "GroupPolicy", "Machine", "Registry.pol"
    )
    reg_pol.unlink(missing_ok=True)
    yield reg_pol
    reg_pol.unlink(missing_ok=True)


@pytest.fixture
def clean_user():
    reg_pol = pathlib.Path(
        os.getenv("SystemRoot"), "System32", "GroupPolicy", "User", "Registry.pol"
    )
    reg_pol.unlink(missing_ok=True)
    yield reg_pol
    reg_pol.unlink(missing_ok=True)


@pytest.fixture(scope="module")
def lgpo_bin():
    sys_dir = pathlib.Path(os.getenv("SystemRoot"), "System32")
    if not (sys_dir / "lgpo.exe").exists():
        zip_file = sys_dir / "lgpo.zip"
        # download lgpo.zip
        log.debug("Downloading LGPO.exe from Microsoft")
        url = "https://download.microsoft.com/download/8/5/C/85C25433-A1B0-4FFA-9429-7E023E7DA8D8/LGPO.zip"
        r = requests.get(url, timeout=60)
        with salt.utils.files.fopen(zip_file, "wb") as f:
            f.write(r.content)
        # extract zip
        log.debug("Extracting LGPO.exe")
        with zipfile.ZipFile(zip_file) as z:
            for file in z.namelist():
                if file.lower().endswith("lgpo.exe"):
                    location = z.extract(file, path=r"C:\Windows\System32")
        # Move lgpo.exe to system32
        log.debug("Placing LGPO.exe in System32")
        lgpo_bin = pathlib.Path(location)
        lgpo_bin = lgpo_bin.rename(sys_dir / lgpo_bin.name.lower())
        yield lgpo_bin
        log.debug("Cleaning up LGPO artifacts")
        zip_file.unlink(missing_ok=True)
        lgpo_bin.unlink(missing_ok=True)
        if (sys_dir / "LGPO_30").exists():
            shutil.rmtree(str(sys_dir / "LGPO_30"))
    else:
        log.debug("LGPO.exe already present")
        yield str(sys_dir / "lgpo.exe")


def test_clear_policy_cache():
    context = {
        "lgpo.policy_definitions": "spongebob",
        "lgpo.policy_resources": "squarepants",
    }
    with patch.dict(win_lgpo.__context__, context):
        assert "lgpo.policy_definitions" in win_lgpo.__context__
        assert "lgpo.policy_resources" in win_lgpo.__context__
        win_lgpo.clear_policy_cache()
        assert "lgpo.policy_definitions" not in win_lgpo.__context__
        assert "lgpo.policy_resources" not in win_lgpo.__context__


@pytest.mark.destructive_test
def test__load_policy_definitions():
    """
    Test that unexpected files in the PolicyDefinitions directory won't
    cause the _load_policy_definitions function to explode
    https://gitlab.com/saltstack/enterprise/lock/issues/3826
    """
    # The PolicyDefinitions directory should only contain ADMX files. We
    # want to make sure the `_load_policy_definitions` function skips non
    # ADMX files in this directory.
    # Create a bogus ADML file in PolicyDefinitions directory
    bogus_fle = os.path.join("c:\\Windows\\PolicyDefinitions", "_bogus.adml")
    cache_dir = os.path.join(win_lgpo.__opts__["cachedir"], "lgpo", "policy_defs")
    try:
        with salt.utils.files.fopen(bogus_fle, "w+") as fh:
            fh.write("<junk></junk>")
        # This function doesn't return anything (None), it just loads
        # the XPath structures into __context__. We're just making sure it
        # doesn't stack trace here
        win_lgpo._load_policy_definitions()
        assert True
    finally:
        # Remove source file
        os.remove(bogus_fle)
        # Remove cached file
        search_string = f"{cache_dir}\\_bogus*.adml"
        for file_name in glob.glob(search_string):
            os.remove(file_name)


def _test_set_computer_policy(lgpo_bin, shell, name, setting, exp_regexes):
    result = win_lgpo.set_computer_policy(name=name, setting=setting)
    assert result is True
    ret = shell.run(
        lgpo_bin,
        "/parse",
        "/m",
        r"C:\Windows\System32\GroupPolicy\Machine\Registry.pol",
    )
    assert ret.returncode == 0
    content = ret.stdout
    # Make sure it's a valid file format
    assert re.search(r"Invalid file format\.", content, re.IGNORECASE) is None
    for exp_regex in exp_regexes:
        match = re.search(exp_regex, content, re.IGNORECASE | re.MULTILINE)
        assert match is not None


def _test_set_user_policy(lgpo_bin, shell, name, setting, exp_regexes):
    result = win_lgpo.set_user_policy(name=name, setting=setting)
    assert result is True
    ret = shell.run(
        lgpo_bin,
        "/parse",
        "/u",
        r"C:\Windows\System32\GroupPolicy\User\Registry.pol",
    )
    assert ret.returncode == 0
    content = ret.stdout
    # Make sure it's a valid file format
    assert re.search(r"Invalid file format\.", content, re.IGNORECASE) is None
    for exp_regex in exp_regexes:
        match = re.search(exp_regex, content, re.IGNORECASE | re.MULTILINE)
        assert match is not None


@pytest.mark.parametrize(
    "name, setting, exp_regexes",
    [
        (
            r"Configure Windows NTP Client",
            "Disabled",
            [
                r"Computer[\s]*Software\\Policies\\Microsoft\\W32time\\Parameters[\s]*NtpServer[\s]*DELETE",
                r"Computer[\s]*Software\\Policies\\Microsoft\\W32time\\Parameters[\s]*Type[\s]*DELETE",
                r"Computer[\s]*Software\\Policies\\Microsoft\\W32time\\TimeProviders\\NtpClient[\s]*CrossSiteSyncFlags[\s]*DELETE",
                r"Computer[\s]*Software\\Policies\\Microsoft\\W32time\\TimeProviders\\NtpClient[\s]*ResolvePeerBackoffMinutes[\s]*DELETE",
                r"Computer[\s]*Software\\Policies\\Microsoft\\W32time\\TimeProviders\\NtpClient[\s]*ResolvePeerBackoffMaxTimes[\s]*DELETE",
                r"Computer[\s]*Software\\Policies\\Microsoft\\W32time\\TimeProviders\\NtpClient[\s]*SpecialPollInterval[\s]*DELETE",
                r"Computer[\s]*Software\\Policies\\Microsoft\\W32time\\TimeProviders\\NtpClient[\s]*EventLogFlags[\s]*DELETE",
            ],
        ),
        (
            "Configure Windows NTP Client",
            {
                "NtpServer": "time.windows.com,0x9",
                "Type": "NT5DS",
                "CrossSiteSyncFlags": 2,
                "ResolvePeerBackoffMinutes": 15,
                "ResolvePeerBackoffMaxTimes": 7,
                "W32TIME_SpecialPollInterval": 3600,
                "W32TIME_NtpClientEventLogFlags": 0,
            },
            [
                r"Computer[\s]*Software\\Policies\\Microsoft\\W32time\\Parameters[\s]*NtpServer[\s]*SZ:time.windows.com,0x9",
                r"Computer[\s]*Software\\Policies\\Microsoft\\W32time\\Parameters[\s]*Type[\s]*SZ:NT5DS",
                r"Computer[\s]*Software\\Policies\\Microsoft\\W32time\\TimeProviders\\NtpClient[\s]*CrossSiteSyncFlags[\s]*DWORD:2",
                r"Computer[\s]*Software\\Policies\\Microsoft\\W32time\\TimeProviders\\NtpClient[\s]*ResolvePeerBackoffMinutes[\s]*DWORD:15",
                r"Computer[\s]*Software\\Policies\\Microsoft\\W32time\\TimeProviders\\NtpClient[\s]*ResolvePeerBackoffMaxTimes[\s]*DWORD:7",
                r"Computer[\s]*Software\\Policies\\Microsoft\\W32time\\TimeProviders\\NtpClient[\s]*SpecialPollInterval[\s]*DWORD:3600",
                r"Computer[\s]*Software\\Policies\\Microsoft\\W32time\\TimeProviders\\NtpClient[\s]*EventLogFlags[\s]*DWORD:0",
            ],
        ),
        (
            "Configure Windows NTP Client",
            "Not Configured",
            [
                r"; Source file:  C:\\Windows\\System32\\GroupPolicy\\Machine\\Registry.pol[\s]*; PARSING COMPLETED."
            ],
        ),
        (
            "RA_Unsolicit",
            "Disabled",
            [
                r"Computer[\s]*Software\\policies\\Microsoft\\Windows NT\\Terminal Services[\s]*fAllowUnsolicited[\s]*DWORD:0",
                r"Computer[\s]*Software\\policies\\Microsoft\\Windows NT\\Terminal Services[\s]*fAllowUnsolicitedFullControl[\s]*DELETE",
                r"Computer[\s]*Software\\policies\\Microsoft\\Windows NT\\Terminal Services\\RAUnsolicit[\s]*\*[\s]*DELETEALLVALUES",
            ],
        ),
        (
            "RA_Unsolicit",
            {
                "Permit remote control of this computer": (
                    "Allow helpers to remotely control the computer"
                ),
                "Helpers": ["administrators", "user1"],
            },
            [
                r"Computer[\s]*Software\\policies\\Microsoft\\Windows NT\\Terminal Services\\RAUnsolicit[\s]*user1[\s]*SZ:user1[\s]*",
                r"Computer[\s]*Software\\policies\\Microsoft\\Windows NT\\Terminal Services\\RAUnsolicit[\s]*administrators[\s]*SZ:administrators[\s]*",
                r"Computer[\s]*Software\\policies\\Microsoft\\Windows NT\\Terminal Services[\s]*fAllowUnsolicited[\s]*DWORD:1",
                r"Computer[\s]*Software\\policies\\Microsoft\\Windows NT\\Terminal Services[\s]*fAllowUnsolicitedFullControl[\s]*DWORD:1",
            ],
        ),
        (
            "RA_Unsolicit",
            "Not Configured",
            [
                r"; Source file:  C:\\Windows\\System32\\GroupPolicy\\Machine\\Registry.pol[\s]*; PARSING COMPLETED."
            ],
        ),
        (
            "Pol_HardenedPaths",
            "Disabled",
            [
                r"Computer[\s]*Software\\policies\\Microsoft\\Windows\\NetworkProvider\\HardenedPaths[\s]*\*[\s]*DELETEALLVALUES"
            ],
        ),
        (
            "Pol_HardenedPaths",
            {
                "Hardened UNC Paths": {
                    r"\\*\NETLOGON": (
                        "RequireMutualAuthentication=1, RequireIntegrity=1"
                    ),
                    r"\\*\SYSVOL": "RequireMutualAuthentication=1, RequireIntegrity=1",
                }
            },
            [
                r"Computer[\s]*Software\\policies\\Microsoft\\Windows\\NetworkProvider\\HardenedPaths[\s]*\\\\\*\\NETLOGON[\s]*SZ:RequireMutualAuthentication=1, RequireIntegrity=1[\s]*",
                r"Computer[\s]*Software\\policies\\Microsoft\\Windows\\NetworkProvider\\HardenedPaths[\s]*\\\\\*\\SYSVOL[\s]*SZ:RequireMutualAuthentication=1, RequireIntegrity=1[\s]*",
            ],
        ),
        (
            "Pol_HardenedPaths",
            "Not Configured",
            [
                r"; Source file:  C:\\Windows\\System32\\GroupPolicy\\Machine\\Registry.pol[\s]*; PARSING COMPLETED."
            ],
        ),
        (
            "TS_CLIENT_CLIPBOARD",
            "Enabled",
            [
                r"Computer[\s]*Software\\Policies\\Microsoft\\Windows NT\\Terminal Services[\s]*fDisableClip[\s]*DWORD:1"
            ],
        ),
        (
            # Use key name until github issue #62732 is fixed
            # "Do not allow Clipboard redirection",
            "TS_CLIENT_CLIPBOARD",
            "Disabled",
            [
                r"Computer[\s]*Software\\Policies\\Microsoft\\Windows NT\\Terminal Services[\s]*fDisableClip[\s]*DWORD:0"
            ],
        ),
        (
            r"Windows Components\Remote Desktop Services\Remote Desktop Session Host\Device and Resource Redirection\Do not allow Clipboard redirection",
            "Not Configured",
            [
                r"; Source file:  C:\\Windows\\System32\\GroupPolicy\\Machine\\Registry.pol[\s]*; PARSING COMPLETED."
            ],
        ),
        # Windows Server 2016+ Policies vvvvvvvvvvvvvvvvvvvvvvvv
        (
            "DisableUXWUAccess",
            "Enabled",
            [
                r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\WindowsUpdate[\s]*SetDisableUXWUAccess[\s]*DWORD:1"
            ],
        ),
        (
            "Remove access to use all Windows Update features",
            "Disabled",
            [
                r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\WindowsUpdate[\s]*SetDisableUXWUAccess[\s]*DWORD:0"
            ],
        ),
        (
            r"ActiveHours",
            {"ActiveHoursStartTime": "8 AM", "ActiveHoursEndTime": "7 PM"},
            [
                r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\WindowsUpdate[\s]*SetActiveHours[\s]*DWORD:1",
                r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\WindowsUpdate[\s]*ActiveHoursStart[\s]*DWORD:8",
                r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\WindowsUpdate[\s]*ActiveHoursEnd[\s]*DWORD:19",
            ],
        ),
        (
            "Turn off auto-restart for updates during active hours",
            "Disabled",
            [
                r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\WindowsUpdate[\s]*SetActiveHours[\s]*DWORD:0",
                r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\WindowsUpdate[\s]*ActiveHoursStart[\s]*DELETE",
                r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\WindowsUpdate[\s]*ActiveHoursEnd[\s]*DELETE",
            ],
        ),
        (
            "Specify settings for optional component installation and component repair",
            "Not Configured",
            [
                r"; Source file:  C:\\Windows\\System32\\GroupPolicy\\Machine\\Registry.pol[\s]*; PARSING COMPLETED."
            ],
        ),
        # Policies with multiple names
        (
            r"Windows Components\Internet Explorer\Internet Control Panel\Security Page\Internet Zone\Access data sources across domains",
            {"Access data sources across domains": "Prompt"},
            [
                r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\CurrentVersion\\Internet Settings\\Zones\\3[\s]*1406[\s]*DWORD:1"
            ],
        ),
        (
            r"Windows Components\Internet Explorer\Internet Control Panel\Security Page\Locked-Down Internet Zone\Access data sources across domains",
            {"Access data sources across domains": "Prompt"},
            [
                r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\CurrentVersion\\Internet Settings\\lockdown_zones\\3[\s]*1406[\s]*DWORD:1"
            ],
        ),
        (
            r"Windows Components\Internet Explorer\Internet Control Panel\Security Page\Intranet Zone\Access data sources across domains",
            {"Access data sources across domains": "Prompt"},
            [
                r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\CurrentVersion\\Internet Settings\\Zones\\1[\s]*1406[\s]*DWORD:1"
            ],
        ),
        (
            r"Windows Components\Internet Explorer\Internet Control Panel\Security Page\Locked-Down Intranet Zone\Access data sources across domains",
            {"Access data sources across domains": "Prompt"},
            [
                r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\CurrentVersion\\Internet Settings\\lockdown_zones\\1[\s]*1406[\s]*DWORD:1"
            ],
        ),
        (
            r"Windows Components\Internet Explorer\Internet Control Panel\Security Page\Local Machine Zone\Access data sources across domains",
            {"Access data sources across domains": "Prompt"},
            [
                r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\CurrentVersion\\Internet Settings\\Zones\\0[\s]*1406[\s]*DWORD:1"
            ],
        ),
        (
            r"Windows Components\Internet Explorer\Internet Control Panel\Security Page\Locked-Down Local Machine Zone\Access data sources across domains",
            {"Access data sources across domains": "Prompt"},
            [
                r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\CurrentVersion\\Internet Settings\\lockdown_zones\\0[\s]*1406[\s]*DWORD:1"
            ],
        ),
        (
            r"Windows Components\Internet Explorer\Internet Control Panel\Security Page\Restricted Sites Zone\Access data sources across domains",
            {"Access data sources across domains": "Prompt"},
            [
                r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\CurrentVersion\\Internet Settings\\Zones\\4[\s]*1406[\s]*DWORD:1"
            ],
        ),
        (
            r"Windows Components\Internet Explorer\Internet Control Panel\Security Page\Locked-Down Restricted Sites Zone\Access data sources across domains",
            {"Access data sources across domains": "Prompt"},
            [
                r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\CurrentVersion\\Internet Settings\\lockdown_zones\\4[\s]*1406[\s]*DWORD:1"
            ],
        ),
        (
            r"Windows Components\Internet Explorer\Internet Control Panel\Security Page\Trusted Sites Zone\Access data sources across domains",
            {"Access data sources across domains": "Prompt"},
            [
                r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\CurrentVersion\\Internet Settings\\Zones\\2[\s]*1406[\s]*DWORD:1"
            ],
        ),
        (
            r"Windows Components\Internet Explorer\Internet Control Panel\Security Page\Locked-Down Trusted Sites Zone\Access data sources across domains",
            {"Access data sources across domains": "Prompt"},
            [
                r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\CurrentVersion\\Internet Settings\\lockdown_zones\\2[\s]*1406[\s]*DWORD:1"
            ],
        ),
    ],
)
@pytest.mark.destructive_test
def test_set_computer_policy(clean_comp, lgpo_bin, shell, name, setting, exp_regexes):
    _test_set_computer_policy(
        lgpo_bin=lgpo_bin,
        shell=shell,
        name=name,
        setting=setting,
        exp_regexes=exp_regexes,
    )


@pytest.mark.parametrize(
    "name, setting, exp_regexes",
    [
        (
            "Specify settings for optional component installation and component repair",
            "Disabled",
            [
                r"Computer[\s]*Software\\Microsoft\\Windows\\CurrentVersion\\Policies\\Servicing[\s]*LocalSourcePath[\s]*DELETE",
                r"Computer[\s]*Software\\Microsoft\\Windows\\CurrentVersion\\Policies\\Servicing[\s]*UseWindowsUpdate[\s]*DELETE",
                r"Computer[\s]*Software\\Microsoft\\Windows\\CurrentVersion\\Policies\\Servicing[\s]*RepairContentServerSource[\s]*DELETE",
            ],
        ),
        (
            "Specify settings for optional component installation and component repair",
            {
                "Alternate source file path": "",
                "Never attempt to download payload from Windows Update": True,
                "CheckBox_SidestepWSUS": False,
            },
            [
                r"Computer[\s]*Software\\Microsoft\\Windows\\CurrentVersion\\Policies\\Servicing[\s]*LocalSourcePath[\s]*EXSZ:",
                r"Computer[\s]*Software\\Microsoft\\Windows\\CurrentVersion\\Policies\\Servicing[\s]*UseWindowsUpdate[\s]*DWORD:2",
                r"Computer[\s]*Software\\Microsoft\\Windows\\CurrentVersion\\Policies\\Servicing[\s]*RepairContentServerSource[\s]*DELETE",
            ],
        ),
        (
            "Specify settings for optional component installation and component repair",
            {
                "Alternate source file path": r"\\some\fake\server",
                "Never attempt to download payload from Windows Update": True,
                "CheckBox_SidestepWSUS": False,
            },
            [
                r"Computer[\s]*Software\\Microsoft\\Windows\\CurrentVersion\\Policies\\Servicing[\s]*LocalSourcePath[\s]*EXSZ:\\\\\\\\some\\\\fake\\\\server",
                r"Computer[\s]*Software\\Microsoft\\Windows\\CurrentVersion\\Policies\\Servicing[\s]*UseWindowsUpdate[\s]*DWORD:2",
                r"Computer[\s]*Software\\Microsoft\\Windows\\CurrentVersion\\Policies\\Servicing[\s]*RepairContentServerSource[\s]*DELETE",
            ],
        ),
    ],
)
@pytest.mark.destructive_test
def test_set_computer_policy_old(
    clean_comp, lgpo_bin, shell, name, setting, exp_regexes, osversion
):
    """
    This tests the policy on Windows Server 2022 and older. Newer versions have
    removed a couple parameters
    """
    if salt.utils.versions.compare(ver1=osversion, oper=">=", ver2="10.0.26100"):
        pytest.skip(f"Test not compatible with {osversion}")

    _test_set_computer_policy(
        lgpo_bin=lgpo_bin,
        shell=shell,
        name=name,
        setting=setting,
        exp_regexes=exp_regexes,
    )


@pytest.mark.parametrize(
    "name, setting, exp_regexes",
    [
        (
            "Specify settings for optional component installation and component repair",
            "Disabled",
            [
                r"Computer[\s]*Software\\Microsoft\\Windows\\CurrentVersion\\Policies\\Servicing[\s]*LocalSourcePath[\s]*DELETE",
            ],
        ),
        (
            "Specify settings for optional component installation and component repair",
            {
                "Alternate source file path": "",
            },
            [
                r"Computer[\s]*Software\\Microsoft\\Windows\\CurrentVersion\\Policies\\Servicing[\s]*LocalSourcePath[\s]*EXSZ:",
            ],
        ),
        (
            "Specify settings for optional component installation and component repair",
            {
                "Alternate source file path": r"\\some\fake\server",
            },
            [
                r"Computer[\s]*Software\\Microsoft\\Windows\\CurrentVersion\\Policies\\Servicing[\s]*LocalSourcePath[\s]*EXSZ:\\\\\\\\some\\\\fake\\\\server",
            ],
        ),
    ],
)
@pytest.mark.destructive_test
def test_set_computer_policy_new(
    clean_comp, lgpo_bin, shell, name, setting, exp_regexes, osversion
):
    """
    This tests the policy on Windows Server 2025. Older versions have additional
    parameters that are set.
    """
    if salt.utils.versions.compare(ver1=osversion, oper="<", ver2="10.0.26100"):
        pytest.skip(f"Test not compatible with {osversion}")
    _test_set_computer_policy(
        lgpo_bin=lgpo_bin,
        shell=shell,
        name=name,
        setting=setting,
        exp_regexes=exp_regexes,
    )


@pytest.mark.parametrize(
    "name, setting, exp_regexes",
    [
        (
            "Point and Print Restrictions",
            {
                "Users can only point and print to these servers": True,
                "Enter fully qualified server names separated by semicolons": (
                    "fakeserver1;fakeserver2"
                ),
                "Users can only point and print to machines in their forest": True,
                "When installing drivers for a new connection": (
                    "Show warning and elevation prompt"
                ),
                "When updating drivers for an existing connection": "Show warning only",
            },
            [
                r"User[\s]*Software\\Policies\\Microsoft\\Windows NT\\Printers"
                r"\\PointAndPrint[\s]*Restricted[\s]*DWORD:1",
                r"User[\s]*Software\\Policies\\Microsoft\\Windows NT\\Printers"
                r"\\PointAndPrint[\s]*TrustedServers[\s]*DWORD:1",
                r"User[\s]*Software\\Policies\\Microsoft\\Windows NT\\Printers"
                r"\\PointAndPrint[\s]*ServerList[\s]*SZ:fakeserver1;fakeserver2",
                r"User[\s]*Software\\Policies\\Microsoft\\Windows NT\\Printers"
                r"\\PointAndPrint[\s]*InForest[\s]*DWORD:1",
                r"User[\s]*Software\\Policies\\Microsoft\\Windows NT\\Printers"
                r"\\PointAndPrint[\s]*NoWarningNoElevationOnInstall[\s]*DWORD:0",
                r"User[\s]*Software\\Policies\\Microsoft\\Windows NT\\Printers"
                r"\\PointAndPrint[\s]*UpdatePromptSettings[\s]*DWORD:1",
            ],
        ),
        (
            "Point and Print Restrictions",
            "Disabled",
            [
                r"User[\s]*Software\\Policies\\Microsoft\\Windows NT\\Printers"
                r"\\PointAndPrint[\s]*Restricted[\s]*DWORD:0",
                r"User[\s]*Software\\Policies\\Microsoft\\Windows NT\\Printers"
                r"\\PointAndPrint[\s]*TrustedServers[\s]*DELETE",
                r"User[\s]*Software\\Policies\\Microsoft\\Windows NT\\Printers"
                r"\\PointAndPrint[\s]*ServerList[\s]*DELETE",
                r"User[\s]*Software\\Policies\\Microsoft\\Windows NT\\Printers"
                r"\\PointAndPrint[\s]*InForest[\s]*DELETE",
                r"User[\s]*Software\\Policies\\Microsoft\\Windows NT\\Printers"
                r"\\PointAndPrint[\s]*NoWarningNoElevationOnInstall[\s]*DELETE",
                r"User[\s]*Software\\Policies\\Microsoft\\Windows NT\\Printers"
                r"\\PointAndPrint[\s]*UpdatePromptSettings[\s]*DELETE",
            ],
        ),
        (
            "Point and Print Restrictions",
            "Not Configured",
            [
                r"; Source file: "
                r" c:\\windows\\system32\\grouppolicy\\user\\registry.pol[\s]*;"
                r" PARSING COMPLETED."
            ],
        ),
    ],
)
@pytest.mark.destructive_test
def test_set_user_policy(clean_user, lgpo_bin, shell, name, setting, exp_regexes):
    _test_set_user_policy(
        lgpo_bin=lgpo_bin,
        shell=shell,
        name=name,
        setting=setting,
        exp_regexes=exp_regexes,
    )


@pytest.mark.destructive_test
def test_set_computer_policy_windows_update(clean_comp, lgpo_bin, shell):
    """
    Test setting/unsetting/changing WindowsUpdate policy
    """
    # Configure Automatic Updates has different options in different builds and
    # releases of Windows, so we'll get the elements and add them if they are
    # present. Newer elements will need to be added manually as they are
    # made available in newer releases of Windows
    result = win_lgpo.get_policy_info(
        policy_name="Configure Automatic Updates",
        policy_class="Machine",
    )
    the_policy = {}
    the_policy_check_enabled = [
        r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\WindowsUpdate\\AU[\s]*NoAutoUpdate[\s]*DWORD:0",
    ]
    the_policy_check_disabled = [
        r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\WindowsUpdate\\AU[\s]*NoAutoUpdate[\s]*DWORD:1",
    ]
    for item in result["policy_elements"]:
        if "Configure automatic updating" in item["element_aliases"]:
            the_policy.update(
                {
                    "Configure automatic updating": (
                        "4 - Auto download and schedule the install"
                    ),
                }
            )
            the_policy_check_enabled.append(
                r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\WindowsUpdate\\AU[\s]*AUOptions[\s]*DWORD:4",
            )
            the_policy_check_disabled.append(
                r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\WindowsUpdate\\AU[\s]*AUOptions[\s]*DELETE",
            )
        elif "Install during automatic maintenance" in item["element_aliases"]:
            the_policy.update({"Install during automatic maintenance": True})
            the_policy_check_enabled.append(
                r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\WindowsUpdate\\AU[\s]*AutomaticMaintenanceEnabled[\s]*DWORD:1\s*",
            )
            the_policy_check_disabled.append(
                r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\WindowsUpdate\\AU[\s]*AutomaticMaintenanceEnabled[\s]*DELETE",
            )
        elif "Scheduled install day" in item["element_aliases"]:
            the_policy.update({"Scheduled install day": "7 - Every Saturday"})
            the_policy_check_enabled.append(
                r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\WindowsUpdate\\AU[\s]*ScheduledInstallDay[\s]*DWORD:7",
            )
            the_policy_check_disabled.append(
                r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\WindowsUpdate\\AU[\s]*ScheduledInstallDay[\s]*DELETE",
            )
        elif "Scheduled install time" in item["element_aliases"]:
            the_policy.update({"Scheduled install time": "17:00"})
            the_policy_check_enabled.append(
                r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\WindowsUpdate\\AU[\s]*ScheduledInstallTime[\s]*DWORD:17",
            )
            the_policy_check_disabled.append(
                r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\WindowsUpdate\\AU[\s]*ScheduledInstallTime[\s]*DELETE",
            )
        elif "Install updates for other Microsoft products" in item["element_aliases"]:
            the_policy.update({"Install updates for other Microsoft products": True})
            the_policy_check_enabled.append(
                r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\WindowsUpdate\\AU[\s]*AllowMUUpdateService[\s]*DWORD:1\s*"
            )
            the_policy_check_disabled.append(
                r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\WindowsUpdate\\AU[\s]*AllowMUUpdateService[\s]*DELETE"
            )
        elif "AutoUpdateSchEveryWeek" in item["element_aliases"]:
            the_policy.update({"AutoUpdateSchEveryWeek": True})
            the_policy_check_enabled.append(
                r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\WindowsUpdate\\AU[\s]*ScheduledInstallEveryWeek[\s]*DWORD:1\s*"
            )
            the_policy_check_disabled.append(
                r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\WindowsUpdate\\AU[\s]*ScheduledInstallEveryWeek[\s]*DELETE"
            )
        elif "First week of the month" in item["element_aliases"]:
            the_policy.update({"First week of the month": True})
            the_policy_check_enabled.append(
                r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\WindowsUpdate\\AU[\s]*ScheduledInstallFirstWeek[\s]*DWORD:1\s*"
            )
            the_policy_check_disabled.append(
                r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\WindowsUpdate\\AU[\s]*ScheduledInstallFirstWeek[\s]*DELETE"
            )
        elif "Second week of the month" in item["element_aliases"]:
            the_policy.update({"Second week of the month": True})
            the_policy_check_enabled.append(
                r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\WindowsUpdate\\AU[\s]*ScheduledInstallSecondWeek[\s]*DWORD:1\s*"
            )
            the_policy_check_disabled.append(
                r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\WindowsUpdate\\AU[\s]*ScheduledInstallSecondWeek[\s]*DELETE"
            )
        elif "Third week of the month" in item["element_aliases"]:
            the_policy.update({"Third week of the month": True})
            the_policy_check_enabled.append(
                r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\WindowsUpdate\\AU[\s]*ScheduledInstallThirdWeek[\s]*DWORD:1\s*"
            )
            the_policy_check_disabled.append(
                r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\WindowsUpdate\\AU[\s]*ScheduledInstallThirdWeek[\s]*DELETE"
            )
        elif "Fourth week of the month" in item["element_aliases"]:
            the_policy.update({"Fourth week of the month": True})
            the_policy_check_enabled.append(
                r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\WindowsUpdate\\AU[\s]*ScheduledInstallFourthWeek[\s]*DWORD:1\s*"
            )
            the_policy_check_disabled.append(
                r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\WindowsUpdate\\AU[\s]*ScheduledInstallFourthWeek[\s]*DELETE"
            )

    # enable Automatic Updates
    _test_set_computer_policy(
        lgpo_bin=lgpo_bin,
        shell=shell,
        name="Configure Automatic Updates",
        setting=the_policy,
        exp_regexes=the_policy_check_enabled,
    )
    clean_comp.unlink(missing_ok=True)
    # disable Configure Automatic Updates
    _test_set_computer_policy(
        lgpo_bin=lgpo_bin,
        shell=shell,
        name="Configure Automatic Updates",
        setting="Disabled",
        exp_regexes=the_policy_check_disabled,
    )
    clean_comp.unlink(missing_ok=True)
    # set Configure Automatic Updates to 'Not Configured'
    _test_set_computer_policy(
        lgpo_bin=lgpo_bin,
        shell=shell,
        name="Configure Automatic Updates",
        setting="Not Configured",
        exp_regexes=[
            r"; Source file: "
            r" c:\\windows\\system32\\grouppolicy\\machine\\registry.pol[\s]*;"
            r" PARSING COMPLETED."
        ],
    )


@pytest.mark.destructive_test
def test_set_computer_policy_multiple_policies(clean_comp, lgpo_bin, shell):
    """
    Tests setting several ADMX policies in succession and validating the
    configuration
    """
    # set one policy
    _test_set_computer_policy(
        lgpo_bin=lgpo_bin,
        shell=shell,
        name="TS_CLIENT_CLIPBOARD",
        setting="Disabled",
        exp_regexes=[
            r"Computer[\s]*Software\\Policies\\Microsoft\\Windows NT"
            r"\\Terminal Services[\s]*fDisableClip[\s]*DWORD:0",
        ],
    )
    # set another policy and make sure the previous policy is OK
    _test_set_computer_policy(
        lgpo_bin=lgpo_bin,
        shell=shell,
        name="RA_Unsolicit",
        setting={
            "Permit remote control of this computer": (
                "Allow helpers to remotely control the computer"
            ),
            "Helpers": ["administrators", "user1"],
        },
        exp_regexes=[
            r"Computer[\s]*Software\\Policies\\Microsoft\\Windows NT"
            r"\\Terminal Services[\s]*fDisableClip[\s]*DWORD:0",
            r"Computer[\s]*Software\\policies\\Microsoft\\Windows NT"
            r"\\Terminal Services\\RAUnsolicit[\s]*user1[\s]*SZ:user1[\s]*",
            r"Computer[\s]*Software\\policies\\Microsoft\\Windows NT"
            r"\\Terminal Services\\RAUnsolicit[\s]*administrators[\s]*SZ:administrators[\s]*",
            r"Computer[\s]*Software\\policies\\Microsoft\\Windows NT"
            r"\\Terminal Services[\s]*fAllowUnsolicited[\s]*DWORD:1",
            r"Computer[\s]*Software\\policies\\Microsoft\\Windows NT"
            r"\\Terminal Services[\s]*fAllowUnsolicitedFullControl[\s]*DWORD:1",
        ],
    )
    # configure automatic updates and make sure everything is still OK
    _test_set_computer_policy(
        lgpo_bin=lgpo_bin,
        shell=shell,
        name="Configure Automatic Updates",
        setting="Disabled",
        exp_regexes=[
            r"Computer[\s]*Software\\Policies\\Microsoft\\Windows NT"
            r"\\Terminal Services[\s]*fDisableClip[\s]*DWORD:0",
            r"Computer[\s]*Software\\policies\\Microsoft\\Windows NT"
            r"\\Terminal Services\\RAUnsolicit[\s]*user1[\s]*SZ:user1[\s]*",
            r"Computer[\s]*Software\\policies\\Microsoft\\Windows NT"
            r"\\Terminal Services\\RAUnsolicit[\s]*administrators[\s]*SZ:administrators[\s]*",
            r"Computer[\s]*Software\\policies\\Microsoft\\Windows NT"
            r"\\Terminal Services[\s]*fAllowUnsolicited[\s]*DWORD:1",
            r"Computer[\s]*Software\\policies\\Microsoft\\Windows NT"
            r"\\Terminal Services[\s]*fAllowUnsolicitedFullControl[\s]*DWORD:1",
            r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\WindowsUpdate"
            r"\\AU[\s]*NoAutoUpdate[\s]*DWORD:1",
            r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\WindowsUpdate"
            r"\\AU[\s]*AUOptions[\s]*DELETE",
            r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\WindowsUpdate"
            r"\\AU[\s]*AutomaticMaintenanceEnabled[\s]*DELETE",
            r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\WindowsUpdate"
            r"\\AU[\s]*ScheduledInstallDay[\s]*DELETE",
            r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\WindowsUpdate"
            r"\\AU[\s]*ScheduledInstallTime[\s]*DELETE",
            r"Computer[\s]*Software\\Policies\\Microsoft\\Windows\\WindowsUpdate"
            r"\\AU[\s]*AllowMUUpdateService[\s]*DELETE",
        ],
    )


def test__encode_xmlns_url():
    """
    Tests the _encode_xmlns_url function.
    Spaces in the xmlns url should be converted to %20
    """
    line = '<policyDefinitionResources xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" revision="1.0" schemaVersion="1.0" xmlns="http://schemas.microsoft.com/GroupPolicy/2006/07/Policysecurity intelligence">'
    result = re.sub(
        r'(.*)(\bxmlns(?::\w+)?)\s*=\s*"([^"]+)"(.*)', win_lgpo._encode_xmlns_url, line
    )
    expected = '<policyDefinitionResources xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" revision="1.0" schemaVersion="1.0" xmlns="http://schemas.microsoft.com/GroupPolicy/2006/07/Policysecurity%20intelligence">'
    assert result == expected



import getpass
import pathlib

import salt.channel.client


def test_file_recv_path(salt_master, salt_minion):
    config = salt_minion.config.copy()
    config["master_uri"] = f"tcp://127.0.0.1:{salt_master.config['ret_port']}"
    keyfile = f".{getpass.getuser()}_key"
    data = b"asdf"
    load_path_list = ["..", "..", "..", keyfile]
    cachedir = salt_master.config["cachedir"]
    assert (pathlib.Path(cachedir) / keyfile).exists()
    assert (pathlib.Path(cachedir) / keyfile).read_bytes() != data
    with salt.channel.client.ReqChannel.factory(config, crypt="aes") as channel:
        load = {
            "cmd": "_file_recv",
            "id": salt_minion.config["id"],
            "path": load_path_list,
            "size": len(data),
            "tok": channel.auth.gen_token(b"salt"),
            "loc": 0,
            "data": b"asdf",
        }
        ret = channel.send(load)
    assert ret is False
    assert (pathlib.Path(cachedir) / keyfile).exists()
    assert (pathlib.Path(cachedir) / keyfile).read_bytes() != data



import subprocess

import salt.utils.platform
from tests.support.mock import patch


def test_linux_distribution():
    """
    Test that when `distro` fails with a `subprocess.CalledProcessError` salt
    returns empty strings as default values.
    """
    distro_name = "Salt"
    distro_version = "1"
    distro_codename = "Awesome"
    with patch("distro.name", return_value=distro_name):
        with patch("distro.version", return_value=distro_version), patch(
            "distro.codename", return_value=distro_codename
        ):
            assert salt.utils.platform.linux_distribution() == (
                distro_name,
                distro_version,
                distro_codename,
            )

        distro_version = ""
        with patch(
            "distro.version",
            side_effect=subprocess.CalledProcessError(returncode=1, cmd=["foo"]),
        ), patch("distro.codename", return_value=distro_codename):
            assert salt.utils.platform.linux_distribution() == (
                distro_name,
                distro_version,
                distro_codename,
            )
        distro_codename = ""
        with patch(
            "distro.version",
            side_effect=subprocess.CalledProcessError(returncode=1, cmd=["foo"]),
        ), patch(
            "distro.codename",
            side_effect=subprocess.CalledProcessError(returncode=1, cmd=["foo"]),
        ):
            assert salt.utils.platform.linux_distribution() == (
                distro_name,
                distro_version,
                distro_codename,
            )



import os

import pytest

import salt.modules.cmdmod as cmd
import salt.modules.opkg as opkg
from tests.support.mock import patch

pytestmark = [
    pytest.mark.skip_if_binaries_missing("stat", "md5sum", "uname"),
]


@pytest.fixture
def configure_loader_modules():
    return {
        opkg: {"__salt__": {"cmd.shell": cmd.shell, "cmd.run_stdout": cmd.run_stdout}},
        cmd: {},
    }


def test_conf_d_path_does_not_exist_not_created_by_restart_check(tmp_path):
    """
    Test when nisysapi ``conf.d.path`` does not exist it should not be created by update restart check
    """
    conf_d_path = tmp_path / "conf.d.path"
    assert not conf_d_path.exists()
    with patch.object(opkg, "NILRT_RESTARTCHECK_STATE_PATH", str(tmp_path)), patch(
        "salt.modules.opkg._get_nisysapi_conf_d_path",
        autospec=True,
        return_value=str(conf_d_path),
    ):
        opkg._update_nilrt_restart_state()

        assert not conf_d_path.exists()


def test_conf_d_path_exists_with_no_files(tmp_path):
    """
    Test when nisysapi ``conf.d.path`` exists and no files exist we should not add any check files
    """
    conf_d_path = tmp_path / "conf.d.path"
    conf_d_path.mkdir(parents=True, exist_ok=True)
    with patch.object(opkg, "NILRT_RESTARTCHECK_STATE_PATH", str(tmp_path)), patch(
        "salt.modules.opkg._get_nisysapi_conf_d_path",
        autospec=True,
        return_value=str(conf_d_path),
    ):
        opkg._update_nilrt_restart_state()

        assert not [
            path
            for path in tmp_path.iterdir()
            if path.suffix in (".timestamp", ".md5sum") and path.stem != "modules.dep"
        ]


@pytest.mark.skip_unless_on_linux(reason="Test requires GNU stat")
def test_conf_d_path_exists_with_files(tmp_path):
    """
    Test when nisysapi ``conf.d.path`` exists with files then we should fingerprint the files
    """
    conf_d_path = tmp_path / "conf.d.path"
    conf_d_path.mkdir(parents=True, exist_ok=True)
    file_one = conf_d_path / "file_one"
    expected_md5sum = f"d41d8cd98f00b204e9800998ecf8427e  {file_one}\n"
    expected_timestamp = "10000\n"
    file_one.touch()
    os.utime(str(file_one), (int(expected_timestamp), int(expected_timestamp)))
    with patch.object(opkg, "NILRT_RESTARTCHECK_STATE_PATH", str(tmp_path)), patch(
        "salt.modules.opkg._get_nisysapi_conf_d_path",
        autospec=True,
        return_value=str(conf_d_path),
    ):
        opkg._update_nilrt_restart_state()

        assert [
            path
            for path in tmp_path.iterdir()
            if path.suffix in (".timestamp", ".md5sum") and path.stem != "modules.dep"
        ]

        timestamp = (
            tmp_path / file_one.with_suffix(suffix=".timestamp").name
        ).read_text()
        md5sum = (tmp_path / file_one.with_suffix(suffix=".md5sum").name).read_text()
        assert timestamp == expected_timestamp
        assert md5sum == expected_md5sum



"""
Set grains describing the minion process.
"""

import os

import salt.utils.platform
import salt.utils.user


def _uid():
    """
    Grain for the minion User ID
    """
    return salt.utils.user.get_uid()


def _username():
    """
    Grain for the minion username
    """
    return salt.utils.user.get_user()


def _gid():
    """
    Grain for the minion Group ID
    """
    return salt.utils.user.get_gid()


def _groupname():
    """
    Grain for the minion groupname
    """
    try:
        return salt.utils.user.get_default_group(_username()) or ""
    except KeyError:
        return ""


def _pid():
    """
    Return the current process pid
    """
    return os.getpid()


def grains():
    """
    Return the grains dictionary
    """
    ret = {
        "username": _username(),
        "groupname": _groupname(),
        "pid": _pid(),
    }

    if not salt.utils.platform.is_windows():
        ret["gid"] = _gid()
        ret["uid"] = _uid()

    return ret



"""
React by calling asynchronous runners
"""

import salt.wheel


def cmd(name, fun=None, arg=(), **kwargs):
    """
    Execute a runner asynchronous:

    USAGE:

    .. code-block:: yaml

        run_cloud:
          wheel.cmd:
            - fun: key.delete
            - match: minion_id
    """
    ret = {"name": name, "changes": {}, "comment": "", "result": True}
    if fun is None:
        fun = name
    client = salt.wheel.WheelClient(__opts__)
    low = {"fun": fun, "arg": arg, "kwargs": kwargs}
    client.cmd_async(low)
    return ret



"""
Module for kubeadm
:maintainer:    Alberto Planas <aplanas@suse.com>
:maturity:      new
:depends:       None
:platform:      Linux
"""

import json
import logging
import re

import salt.utils.files
from salt.exceptions import CommandExecutionError

ADMIN_CFG = "/etc/kubernetes/admin.conf"

log = logging.getLogger(__name__)

__virtualname__ = "kubeadm"

# Define not exported variables from Salt, so this can be imported as
# a normal module
try:
    __salt__
except NameError:
    __salt__ = {}


def _api_server_endpoint(config=None):
    """
    Return the API server endpoint
    """
    config = config if config else ADMIN_CFG
    endpoint = None
    try:
        with salt.utils.files.fopen(config, "r") as fp_:
            endpoint = re.search(
                r"^\s*server: https?://(.*)$", fp_.read(), re.MULTILINE
            ).group(1)
    # pylint:disable=broad-except
    except Exception:
        # Any error or exception is mapped to None
        pass
    return endpoint


def _token(create_if_needed=False):
    """
    Return a valid bootstrap token
    """
    tokens = token_list()
    if not tokens and create_if_needed:
        token_create(description="Token created by kubeadm salt module")
        tokens = token_list()
    # We expect that the token is valid for authentication and signing
    return tokens[0]["token"] if tokens else None


def _discovery_token_ca_cert_hash():
    cmd = [
        "openssl",
        "x509",
        "-pubkey",
        "-in",
        "/etc/kubernetes/pki/ca.crt",
        "|",
        "openssl",
        "rsa",
        "-pubin",
        "-outform",
        "der",
        "2>/dev/null",
        "|",
        "openssl",
        "dgst",
        "-sha256",
        "-hex",
        "|",
        "sed",
        "'s/^.* //'",
    ]
    result = __salt__["cmd.run_all"](" ".join(cmd), python_shell=True)
    if result["retcode"]:
        raise CommandExecutionError(result["stderr"])

    return "sha256:{}".format(result["stdout"])


def join_params(create_if_needed=False):
    """
    .. versionadded:: 3001

    Return the parameters required for joining into the cluster

    create_if_needed
       If the token bucket is empty and this parameter is True, a new
       token will be created.

    CLI Example:

    .. code-block:: bash

       salt '*' kubeadm.join_params
       salt '*' kubeadm.join_params create_if_needed=True

    """

    params = {
        "api-server-endpoint": _api_server_endpoint(),
        "token": _token(create_if_needed),
        "discovery-token-ca-cert-hash": _discovery_token_ca_cert_hash(),
    }
    return params


def version(kubeconfig=None, rootfs=None):
    """
    .. versionadded:: 3001

    Return the version of kubeadm

    kubeconfig
       The kubeconfig file to use when talking to the cluster. The
       default values in /etc/kubernetes/admin.conf

    rootfs
       The path to the real host root filesystem

    CLI Example:

    .. code-block:: bash

       salt '*' kubeadm.version

    """
    cmd = ["kubeadm", "version"]

    parameters = [("kubeconfig", kubeconfig), ("rootfs", rootfs)]
    for parameter, value in parameters:
        if value:
            cmd.extend([f"--{parameter}", str(value)])

    cmd.extend(["--output", "json"])

    return json.loads(__salt__["cmd.run_stdout"](cmd))


def _cmd(cmd):
    """Utility function to run commands."""
    result = __salt__["cmd.run_all"](cmd)
    if result["retcode"]:
        raise CommandExecutionError(result["stderr"])
    return result["stdout"]


def token_create(
    token=None,
    config=None,
    description=None,
    groups=None,
    ttl=None,
    usages=None,
    kubeconfig=None,
    rootfs=None,
):
    """
    .. versionadded:: 3001

    Create bootstrap tokens on the server

    token
       Token to write, if None one will be generated. The token must
       match a regular expression, that by default is
       [a-z0-9]{6}.[a-z0-9]{16}

    config
       Path to kubeadm configuration file

    description
       A human friendly description of how this token is used

    groups
       List of extra groups that this token will authenticate, default
       to ['system:bootstrappers:kubeadm:default-node-token']

    ttl
       The duration defore the token is automatically deleted (1s, 2m,
       3h). If set to '0' the token will never expire. Default value
       is 24h0m0s

    usages
       Describes the ways in which this token can be used. The default
       value is ['signing', 'authentication']

    kubeconfig
       The kubeconfig file to use when talking to the cluster. The
       default values in /etc/kubernetes/admin.conf

    rootfs
       The path to the real host root filesystem

    CLI Example:

    .. code-block:: bash

       salt '*' kubeadm.token_create
       salt '*' kubeadm.token_create a1b2c.0123456789abcdef
       salt '*' kubeadm.token_create ttl='6h'
       salt '*' kubeadm.token_create usages="['signing']"

    """
    cmd = ["kubeadm", "token", "create"]
    if token:
        cmd.append(token)

    parameters = [
        ("config", config),
        ("description", description),
        ("groups", groups),
        ("ttl", ttl),
        ("usages", usages),
        ("kubeconfig", kubeconfig),
        ("rootfs", rootfs),
    ]
    for parameter, value in parameters:
        if value:
            if parameter in ("groups", "usages"):
                cmd.extend([f"--{parameter}", json.dumps(value)])
            else:
                cmd.extend([f"--{parameter}", str(value)])

    return _cmd(cmd)


def token_delete(token, kubeconfig=None, rootfs=None):
    """
    .. versionadded:: 3001

    Delete bootstrap tokens on the server

    token
       Token to write, if None one will be generated. The token must
       match a regular expression, that by default is
       [a-z0-9]{6}.[a-z0-9]{16}

    kubeconfig
       The kubeconfig file to use when talking to the cluster. The
       default values in /etc/kubernetes/admin.conf

    rootfs
       The path to the real host root filesystem

    CLI Example:

    .. code-block:: bash

       salt '*' kubeadm.token_delete a1b2c
       salt '*' kubeadm.token_create a1b2c.0123456789abcdef

    """
    cmd = ["kubeadm", "token", "delete", token]

    parameters = [("kubeconfig", kubeconfig), ("rootfs", rootfs)]
    for parameter, value in parameters:
        if value:
            cmd.extend([f"--{parameter}", str(value)])

    return bool(_cmd(cmd))


def token_generate(kubeconfig=None, rootfs=None):
    """
    .. versionadded:: 3001

    Generate and return a bootstrap token, but do not create it on the
    server

    kubeconfig
       The kubeconfig file to use when talking to the cluster. The
       default values in /etc/kubernetes/admin.conf

    rootfs
       The path to the real host root filesystem

    CLI Example:

    .. code-block:: bash

       salt '*' kubeadm.token_generate

    """
    cmd = ["kubeadm", "token", "generate"]

    parameters = [("kubeconfig", kubeconfig), ("rootfs", rootfs)]
    for parameter, value in parameters:
        if value:
            cmd.extend([f"--{parameter}", str(value)])

    return _cmd(cmd)


def token_list(kubeconfig=None, rootfs=None):
    """
    .. versionadded:: 3001

    List bootstrap tokens on the server

    kubeconfig
       The kubeconfig file to use when talking to the cluster. The
       default values in /etc/kubernetes/admin.conf

    rootfs
       The path to the real host root filesystem

    CLI Example:

    .. code-block:: bash

       salt '*' kubeadm.token_list

    """
    cmd = ["kubeadm", "token", "list"]

    parameters = [("kubeconfig", kubeconfig), ("rootfs", rootfs)]
    for parameter, value in parameters:
        if value:
            cmd.extend([f"--{parameter}", str(value)])

    lines = _cmd(cmd).splitlines()

    tokens = []
    if lines:
        # Find the header and parse it.  We do not need to validate
        # the content, as the regex will take care of future changes.
        header = lines.pop(0)
        header = [i.lower() for i in re.findall(r"(\w+(?:\s\w+)*)", header)]

        for line in lines:
            # TODO(aplanas): descriptions with multiple spaces can
            # break the parser.
            values = re.findall(r"(\S+(?:\s\S+)*)", line)
            if len(header) != len(values):
                log.error("Error parsing line: '%s'", line)
                continue
            tokens.append({key: value for key, value in zip(header, values)})
    return tokens


def alpha_certs_renew(rootfs=None):
    """
    .. versionadded:: 3001

    Renews certificates for a Kubernetes cluster

    rootfs
       The path to the real host root filesystem

    CLI Example:

    .. code-block:: bash

       salt '*' kubeadm.alpha_certs_renew

    """
    cmd = ["kubeadm", "alpha", "certs", "renew"]

    parameters = [("rootfs", rootfs)]
    for parameter, value in parameters:
        if value:
            cmd.extend([f"--{parameter}", str(value)])

    return _cmd(cmd)


def alpha_kubeconfig_user(
    client_name,
    apiserver_advertise_address=None,
    apiserver_bind_port=None,
    cert_dir=None,
    org=None,
    token=None,
    rootfs=None,
):
    """
    .. versionadded:: 3001

    Outputs a kubeconfig file for an additional user

    client_name
       The name of the user. It will be used as the CN if client
       certificates are created

    apiserver_advertise_address
       The IP address the API server is accessible on

    apiserver_bind_port
       The port the API server is accessible on (default 6443)

    cert_dir
       The path where certificates are stored (default
       "/etc/kubernetes/pki")

    org
       The organization of the client certificate

    token
       The token that show be used as the authentication mechanism for
       this kubeconfig, instead of client certificates

    rootfs
       The path to the real host root filesystem

    CLI Example:

    .. code-block:: bash

       salt '*' kubeadm.alpha_kubeconfig_user client_name=user

    """
    cmd = ["kubeadm", "alpha", "kubeconfig", "user", "--client-name", client_name]

    parameters = [
        ("apiserver-advertise-address", apiserver_advertise_address),
        ("apiserver-bind-port", apiserver_bind_port),
        ("cert-dir", cert_dir),
        ("org", org),
        ("token", token),
        ("rootfs", rootfs),
    ]
    for parameter, value in parameters:
        if value:
            cmd.extend([f"--{parameter}", str(value)])

    return _cmd(cmd)


def alpha_kubelet_config_download(kubeconfig=None, kubelet_version=None, rootfs=None):
    """
    .. versionadded:: 3001

    Downloads the kubelet configuration from the cluster ConfigMap
    kubelet-config-1.X

    kubeconfig
       The kubeconfig file to use when talking to the cluster. The
       default values in /etc/kubernetes/admin.conf

    kubelet_version
       The desired version for the kubelet

    rootfs
       The path to the real host root filesystem

    CLI Example:

    .. code-block:: bash

       salt '*' kubeadm.alpha_kubelet_config_download
       salt '*' kubeadm.alpha_kubelet_config_download kubelet_version='1.14.0'

    """
    cmd = ["kubeadm", "alpha", "kubelet", "config", "download"]

    parameters = [
        ("kubeconfig", kubeconfig),
        ("kubelet-version", kubelet_version),
        ("rootfs", rootfs),
    ]
    for parameter, value in parameters:
        if value:
            cmd.extend([f"--{parameter}", str(value)])

    return _cmd(cmd)


def alpha_kubelet_config_enable_dynamic(
    node_name, kubeconfig=None, kubelet_version=None, rootfs=None
):
    """
    .. versionadded:: 3001

    Enables or updates dynamic kubelet configuration for a node

    node_name
       Name of the node that should enable the dynamic kubelet
       configuration

    kubeconfig
       The kubeconfig file to use when talking to the cluster. The
       default values in /etc/kubernetes/admin.conf

    kubelet_version
       The desired version for the kubelet

    rootfs
       The path to the real host root filesystem

    CLI Example:

    .. code-block:: bash

       salt '*' kubeadm.alpha_kubelet_config_enable_dynamic node-1

    """
    cmd = [
        "kubeadm",
        "alpha",
        "kubelet",
        "config",
        "enable-dynamic",
        "--node-name",
        node_name,
    ]

    parameters = [
        ("kubeconfig", kubeconfig),
        ("kubelet-version", kubelet_version),
        ("rootfs", rootfs),
    ]
    for parameter, value in parameters:
        if value:
            cmd.extend([f"--{parameter}", str(value)])

    return _cmd(cmd)


def alpha_selfhosting_pivot(
    cert_dir=None,
    config=None,
    kubeconfig=None,
    store_certs_in_secrets=False,
    rootfs=None,
):
    """
    .. versionadded:: 3001

    Converts a static Pod-hosted control plane into a selt-hosted one

    cert_dir
       The path where certificates are stored (default
       "/etc/kubernetes/pki")

    config
       Path to kubeadm configuration file

    kubeconfig
       The kubeconfig file to use when talking to the cluster. The
       default values in /etc/kubernetes/admin.conf

    store_certs_in_secrets
       Enable storing certs in secrets

    rootfs
       The path to the real host root filesystem

    CLI Example:

    .. code-block:: bash

       salt '*' kubeadm.alpha_selfhost_pivot

    """
    cmd = ["kubeadm", "alpha", "selfhosting", "pivot", "--force"]

    if store_certs_in_secrets:
        cmd.append("--store-certs-in-secrets")

    parameters = [
        ("cert-dir", cert_dir),
        ("config", config),
        ("kubeconfig", kubeconfig),
        ("rootfs", rootfs),
    ]
    for parameter, value in parameters:
        if value:
            cmd.extend([f"--{parameter}", str(value)])

    return _cmd(cmd)


def config_images_list(
    config=None,
    feature_gates=None,
    kubernetes_version=None,
    kubeconfig=None,
    rootfs=None,
):
    """
    .. versionadded:: 3001

    Print a list of images kubeadm will use

    config
       Path to kubeadm configuration file

    feature_gates
       A set of key=value pairs that describe feature gates for
       various features

    kubernetes_version
       Choose a specifig Kubernetes version for the control plane
       (default "stable-1")

    kubeconfig
       The kubeconfig file to use when talking to the cluster. The
       default values in /etc/kubernetes/admin.conf

    rootfs
       The path to the real host root filesystem

    CLI Example:

    .. code-block:: bash

       salt '*' kubeadm.config_images_list

    """
    cmd = ["kubeadm", "config", "images", "list"]

    parameters = [
        ("config", config),
        ("feature-gates", feature_gates),
        ("kubernetes-version", kubernetes_version),
        ("kubeconfig", kubeconfig),
        ("rootfs", rootfs),
    ]
    for parameter, value in parameters:
        if value:
            cmd.extend([f"--{parameter}", str(value)])

    return _cmd(cmd).splitlines()


def config_images_pull(
    config=None,
    cri_socket=None,
    feature_gates=None,
    kubernetes_version=None,
    kubeconfig=None,
    rootfs=None,
):
    """
    .. versionadded:: 3001

    Pull images used by kubeadm

    config
       Path to kubeadm configuration file

    cri_socket
       Path to the CRI socket to connect

    feature_gates
       A set of key=value pairs that describe feature gates for
       various features

    kubernetes_version
       Choose a specifig Kubernetes version for the control plane
       (default "stable-1")

    kubeconfig
       The kubeconfig file to use when talking to the cluster. The
       default values in /etc/kubernetes/admin.conf

    rootfs
       The path to the real host root filesystem

    CLI Example:

    .. code-block:: bash

       salt '*' kubeadm.config_images_pull

    """
    cmd = ["kubeadm", "config", "images", "pull"]

    parameters = [
        ("config", config),
        ("cri-socket", cri_socket),
        ("feature-gates", feature_gates),
        ("kubernetes-version", kubernetes_version),
        ("kubeconfig", kubeconfig),
        ("rootfs", rootfs),
    ]
    for parameter, value in parameters:
        if value:
            cmd.extend([f"--{parameter}", str(value)])

    prefix = "[config/images] Pulled "
    return [(line.replace(prefix, "")) for line in _cmd(cmd).splitlines()]


def config_migrate(old_config, new_config=None, kubeconfig=None, rootfs=None):
    """
    .. versionadded:: 3001

    Read an older version of the kubeadm configuration API types from
    a file, and output the similar config object for the newer version

    old_config
       Path to the kubeadm config file that is usin the old API
       version and should be converted

    new_config
       Path to the resulting equivalent kubeadm config file using the
       new API version. If not specified the output will be returned

    kubeconfig
       The kubeconfig file to use when talking to the cluster. The
       default values in /etc/kubernetes/admin.conf

    rootfs
       The path to the real host root filesystem

    CLI Example:

    .. code-block:: bash

       salt '*' kubeadm.config_migrate /oldconfig.cfg

    """
    cmd = ["kubeadm", "config", "migrate", "--old-config", old_config]

    parameters = [
        ("new-config", new_config),
        ("kubeconfig", kubeconfig),
        ("rootfs", rootfs),
    ]
    for parameter, value in parameters:
        if value:
            cmd.extend([f"--{parameter}", str(value)])

    return _cmd(cmd)


def config_print_init_defaults(component_configs=None, kubeconfig=None, rootfs=None):
    """
    .. versionadded:: 3001

    Return default init configuration, that can be used for 'kubeadm
    init'

    component_config
       A comma-separated list for component config API object to print
       the default values for (valid values: KubeProxyConfiguration,
       KubeletConfiguration)

    kubeconfig
       The kubeconfig file to use when talking to the cluster. The
       default values in /etc/kubernetes/admin.conf

    rootfs
       The path to the real host root filesystem

    CLI Example:

    .. code-block:: bash

       salt '*' kubeadm.config_print_init_defaults

    """
    cmd = ["kubeadm", "config", "print", "init-defaults"]

    parameters = [
        ("component-configs", component_configs),
        ("kubeconfig", kubeconfig),
        ("rootfs", rootfs),
    ]
    for parameter, value in parameters:
        if value:
            cmd.extend([f"--{parameter}", str(value)])

    return _cmd(cmd)


def config_print_join_defaults(component_configs=None, kubeconfig=None, rootfs=None):
    """
    .. versionadded:: 3001

    Return default join configuration, that can be used for 'kubeadm
    join'

    component_config
       A comma-separated list for component config API object to print
       the default values for (valid values: KubeProxyConfiguration,
       KubeletConfiguration)

    kubeconfig
       The kubeconfig file to use when talking to the cluster. The
       default values in /etc/kubernetes/admin.conf

    rootfs
       The path to the real host root filesystem

    CLI Example:

    .. code-block:: bash

       salt '*' kubeadm.config_print_join_defaults

    """
    cmd = ["kubeadm", "config", "print", "join-defaults"]

    parameters = [
        ("component-configs", component_configs),
        ("kubeconfig", kubeconfig),
        ("rootfs", rootfs),
    ]
    for parameter, value in parameters:
        if value:
            cmd.extend([f"--{parameter}", str(value)])

    return _cmd(cmd)


def config_upload_from_file(config, kubeconfig=None, rootfs=None):
    """
    .. versionadded:: 3001

    Upload a configuration file to the in-cluster ConfigMap for
    kubeadm configuration

    config
       Path to a kubeadm configuration file

    kubeconfig
       The kubeconfig file to use when talking to the cluster. The
       default values in /etc/kubernetes/admin.conf

    rootfs
       The path to the real host root filesystem

    CLI Example:

    .. code-block:: bash

       salt '*' kubeadm.config_upload_from_file /config.cfg

    """
    cmd = ["kubeadm", "config", "upload", "from-file", "--config", config]

    parameters = [("kubeconfig", kubeconfig), ("rootfs", rootfs)]
    for parameter, value in parameters:
        if value:
            cmd.extend([f"--{parameter}", str(value)])

    return _cmd(cmd)


def config_upload_from_flags(
    apiserver_advertise_address=None,
    apiserver_bind_port=None,
    apiserver_cert_extra_sans=None,
    cert_dir=None,
    cri_socket=None,
    feature_gates=None,
    kubernetes_version=None,
    node_name=None,
    pod_network_cidr=None,
    service_cidr=None,
    service_dns_domain=None,
    kubeconfig=None,
    rootfs=None,
):
    """
    .. versionadded:: 3001

    Create the in-cluster configuration file for the first time using
    flags

    apiserver_advertise_address
       The IP address the API server will advertise it's listening on

    apiserver_bind_port
       The port the API server is accessible on (default 6443)

    apiserver_cert_extra_sans
       Optional extra Subject Alternative Names (SANs) to use for the
       API Server serving certificate

    cert_dir
       The path where to save and store the certificates (default
       "/etc/kubernetes/pki")

    cri_socket
       Path to the CRI socket to connect

    feature_gates
       A set of key=value pairs that describe feature gates for
       various features

    kubernetes_version
       Choose a specifig Kubernetes version for the control plane
       (default "stable-1")

    node_name
       Specify the node name

    pod_network_cidr
       Specify range of IP addresses for the pod network

    service_cidr
       Use alternative range of IP address for service VIPs (default
       "10.96.0.0/12")

    service_dns_domain
       Use alternative domain for services (default "cluster.local")

    kubeconfig
       The kubeconfig file to use when talking to the cluster. The
       default values in /etc/kubernetes/admin.conf

    rootfs
       The path to the real host root filesystem

    CLI Example:

    .. code-block:: bash

       salt '*' kubeadm.config_upload_from_flags

    """
    cmd = ["kubeadm", "config", "upload", "from-flags"]

    parameters = [
        ("apiserver-advertise-address", apiserver_advertise_address),
        ("apiserver-bind-port", apiserver_bind_port),
        ("apiserver-cert-extra-sans", apiserver_cert_extra_sans),
        ("cert-dir", cert_dir),
        ("cri-socket", cri_socket),
        ("feature-gates", feature_gates),
        ("kubernetes-version", kubernetes_version),
        ("node-name", node_name),
        ("pod-network-cidr", pod_network_cidr),
        ("service-cidr", service_cidr),
        ("service-dns-domain", service_dns_domain),
        ("kubeconfig", kubeconfig),
        ("rootfs", rootfs),
    ]
    for parameter, value in parameters:
        if value:
            cmd.extend([f"--{parameter}", str(value)])

    return _cmd(cmd)


def config_view(kubeconfig=None, rootfs=None):
    """
    .. versionadded:: 3001

    View the kubeadm configuration stored inside the cluster

    kubeconfig
       The kubeconfig file to use when talking to the cluster. The
       default values in /etc/kubernetes/admin.conf

    rootfs
       The path to the real host root filesystem

    CLI Example:

    .. code-block:: bash

       salt '*' kubeadm.config_view

    """
    cmd = ["kubeadm", "config", "view"]

    parameters = [("kubeconfig", kubeconfig), ("rootfs", rootfs)]
    for parameter, value in parameters:
        if value:
            cmd.extend([f"--{parameter}", str(value)])

    return _cmd(cmd)


def init(
    apiserver_advertise_address=None,
    apiserver_bind_port=None,
    apiserver_cert_extra_sans=None,
    cert_dir=None,
    certificate_key=None,
    control_plane_endpoint=None,
    config=None,
    cri_socket=None,
    experimental_upload_certs=False,
    upload_certs=False,
    feature_gates=None,
    ignore_preflight_errors=None,
    image_repository=None,
    kubernetes_version=None,
    node_name=None,
    pod_network_cidr=None,
    service_cidr=None,
    service_dns_domain=None,
    skip_certificate_key_print=False,
    skip_phases=None,
    skip_token_print=False,
    token=None,
    token_ttl=None,
    rootfs=None,
):
    """
    .. versionadded:: 3001

    Command to set up the Kubernetes control plane

    apiserver_advertise_address
       The IP address the API server will advertise it's listening on

    apiserver_bind_port
       The port the API server is accessible on (default 6443)

    apiserver_cert_extra_sans
       Optional extra Subject Alternative Names (SANs) to use for the
       API Server serving certificate

    cert_dir
       The path where to save and store the certificates (default
       "/etc/kubernetes/pki")

    certificate_key
       Key used to encrypt the control-plane certificates in the
       kubeadm-certs Secret

    config
       Path to a kubeadm configuration file

    control_plane_endpoint
       Specify a stable IP address or DNS name for the control plane

    cri_socket
       Path to the CRI socket to connect

    experimental_upload_certs
       Upload control-plane certificate to the kubeadm-certs Secret. ( kubeadm version =< 1.16 )

    upload_certs
       Upload control-plane certificate to the kubeadm-certs Secret. ( kubeadm version > 1.16 )

    feature_gates
       A set of key=value pairs that describe feature gates for
       various features

    ignore_preflight_errors
       A list of checks whose errors will be shown as warnings

    image_repository
       Choose a container registry to pull control plane images from

    kubernetes_version
       Choose a specifig Kubernetes version for the control plane
       (default "stable-1")

    node_name
       Specify the node name

    pod_network_cidr
       Specify range of IP addresses for the pod network

    service_cidr
       Use alternative range of IP address for service VIPs (default
       "10.96.0.0/12")

    service_dns_domain
       Use alternative domain for services (default "cluster.local")

    skip_certificate_key_print
       Don't print the key used to encrypt the control-plane
       certificates

    skip_phases
       List of phases to be skipped

    skip_token_print
       Skip printing of the default bootstrap token generated by
       'kubeadm init'

    token
       The token to use for establishing bidirectional trust between
       nodes and control-plane nodes. The token must match a regular
       expression, that by default is [a-z0-9]{6}.[a-z0-9]{16}

    token_ttl
       The duration defore the token is automatically deleted (1s, 2m,
       3h). If set to '0' the token will never expire. Default value
       is 24h0m0s

    rootfs
       The path to the real host root filesystem

    CLI Example:

    .. code-block:: bash

       salt '*' kubeadm.init pod_network_cidr='10.244.0.0/16'

    """
    cmd = ["kubeadm", "init"]

    if experimental_upload_certs:
        cmd.append("--experimental-upload-certs")
    if upload_certs:
        cmd.append("--upload-certs")
    if skip_certificate_key_print:
        cmd.append("--skip-certificate-key-print")
    if skip_token_print:
        cmd.append("--skip-token-print")

    parameters = [
        ("apiserver-advertise-address", apiserver_advertise_address),
        ("apiserver-bind-port", apiserver_bind_port),
        ("apiserver-cert-extra-sans", apiserver_cert_extra_sans),
        ("cert-dir", cert_dir),
        ("certificate-key", certificate_key),
        ("config", config),
        ("control-plane-endpoint", control_plane_endpoint),
        ("cri-socket", cri_socket),
        ("feature-gates", feature_gates),
        ("ignore-preflight-errors", ignore_preflight_errors),
        ("image-repository", image_repository),
        ("kubernetes-version", kubernetes_version),
        ("node-name", node_name),
        ("pod-network-cidr", pod_network_cidr),
        ("service-cidr", service_cidr),
        ("service-dns-domain", service_dns_domain),
        ("skip-phases", skip_phases),
        ("token", token),
        ("token-ttl", token_ttl),
        ("rootfs", rootfs),
    ]
    for parameter, value in parameters:
        if value:
            cmd.extend([f"--{parameter}", str(value)])

    return _cmd(cmd)


# TODO(aplanas):
#   * init_phase_addon_all
#   * init_phase_addon_coredns
#   * init_phase_addon_kube_proxy
#   * init_phase_bootstrap_token
#   * init_phase_certs_all
#   * init_phase_certs_apiserver
#   * init_phase_certs_apiserver_etcd_client
#   * init_phase_certs_apiserver_kubelet_client
#   * init_phase_certs_ca
#   * init_phase_certs_etcd_ca
#   * init_phase_certs_etcd_healthcheck_client
#   * init_phase_certs_etcd_peer
#   * init_phase_certs_etcd_server
#   * init_phase_certs_front_proxy_ca
#   * init_phase_certs_front_proxy_client
#   * init_phase_certs_sa
#   * init_phase_control_plane_all
#   * init_phase_control_plane_apiserver
#   * init_phase_control_plane_controller_manager
#   * init_phase_control_plane_scheduler
#   * init_phase_etcd_local
#   * init_phase_kubeconfig_admin
#   * init_phase_kubeconfig_all
#   * init_phase_kubeconfig_controller_manager
#   * init_phase_kubeconfig_kubelet
#   * init_phase_kubeconfig_scheduler
#   * init_phase_kubelet_start
#   * init_phase_mark_control_plane
#   * init_phase_preflight
#   * init_phase_upload_certs
#   * init_phase_upload_config_all
#   * init_phase_upload_config_kuneadm
#   * init_phase_upload_config_kubelet


def join(
    api_server_endpoint=None,
    apiserver_advertise_address=None,
    apiserver_bind_port=None,
    certificate_key=None,
    config=None,
    cri_socket=None,
    discovery_file=None,
    discovery_token=None,
    discovery_token_ca_cert_hash=None,
    discovery_token_unsafe_skip_ca_verification=False,
    experimental_control_plane=False,
    control_plane=False,
    ignore_preflight_errors=None,
    node_name=None,
    skip_phases=None,
    tls_bootstrap_token=None,
    token=None,
    rootfs=None,
):
    """
    .. versionadded:: 3001

    Command to join to an existing cluster

    api_server_endpoint
       IP address or domain name and port of the API Server

    apiserver_advertise_address
       If the node should host a new control plane instance, the IP
       address the API Server will advertise it's listening on

    apiserver_bind_port
       If the node should host a new control plane instance, the port
       the API Server to bind to (default 6443)

    certificate_key
       Use this key to decrypt the certificate secrets uploaded by
       init

    config
       Path to a kubeadm configuration file

    cri_socket
       Path to the CRI socket to connect

    discovery_file
       For file-based discovery, a file or URL from which to load
       cluster information

    discovery_token
       For token-based discovery, the token used to validate cluster
       information fetched from the API Server

    discovery_token_ca_cert_hash
       For token-based discovery, validate that the root CA public key
       matches this hash (format: "<type>:<value>")

    discovery_token_unsafe_skip_ca_verification
       For token-based discovery, allow joining without
       'discovery-token-ca-cert-hash' pinning

    experimental_control_plane
       Create a new control plane instance on this node (kubeadm version =< 1.16)

    control_plane
       Create a new control plane instance on this node (kubeadm version > 1.16)

    ignore_preflight_errors
       A list of checks whose errors will be shown as warnings

    node_name
       Specify the node name

    skip_phases
       List of phases to be skipped

    tls_bootstrap_token
       Specify the token used to temporarily authenticate with the
       Kubernetes Control Plane while joining the node

    token
       Use this token for both discovery-token and tls-bootstrap-token
       when those values are not provided

    rootfs
       The path to the real host root filesystem

    CLI Example:

    .. code-block:: bash

       salt '*' kubeadm.join 10.160.65.165:6443 token='token'

    """
    cmd = ["kubeadm", "join"]

    if api_server_endpoint:
        cmd.append(api_server_endpoint)
    if discovery_token_unsafe_skip_ca_verification:
        cmd.append("--discovery-token-unsafe-skip-ca-verification")
    if experimental_control_plane:
        cmd.append("--experimental-control-plane")
    if control_plane:
        cmd.append("--control-plane")

    parameters = [
        ("apiserver-advertise-address", apiserver_advertise_address),
        ("apiserver-bind-port", apiserver_bind_port),
        ("certificate-key", certificate_key),
        ("config", config),
        ("cri-socket", cri_socket),
        ("discovery-file", discovery_file),
        ("discovery-token", discovery_token),
        ("discovery-token-ca-cert-hash", discovery_token_ca_cert_hash),
        ("ignore-preflight-errors", ignore_preflight_errors),
        ("node-name", node_name),
        ("skip-phases", skip_phases),
        ("tls-bootstrap-token", tls_bootstrap_token),
        ("token", token),
        ("rootfs", rootfs),
    ]
    for parameter, value in parameters:
        if value:
            cmd.extend([f"--{parameter}", str(value)])

    return _cmd(cmd)


# TODO(aplanas):
#   * join_phase_control_plane_join_all
#   * join_phase_control_plane_join_etcd
#   * join_phase_control_plane_join_mark_control_plane
#   * join_phase_control_plane_join_update_status
#   * join_phase_control_plane_prepare_all
#   * join_phase_control_plane_prepare_certs
#   * join_phase_control_plane_prepare_control_plane
#   * join_phase_control_plane_prepare_download_certs
#   * join_phase_control_plane_prepare_kubeconfig
#   * join_phase_kubelet_start
#   * join_phase_preflight


def reset(
    cert_dir=None,
    cri_socket=None,
    ignore_preflight_errors=None,
    kubeconfig=None,
    rootfs=None,
):
    """
    .. versionadded:: 3001

    Revert any changes made to this host by 'kubeadm init' or 'kubeadm
    join'

    cert_dir
       The path to the directory where the certificates are stored
       (default "/etc/kubernetes/pki")

    cri_socket
       Path to the CRI socket to connect

    ignore_preflight_errors
       A list of checks whose errors will be shown as warnings

    kubeconfig
       The kubeconfig file to use when talking to the cluster. The
       default values in /etc/kubernetes/admin.conf

    rootfs
       The path to the real host root filesystem

    CLI Example:

    .. code-block:: bash

       salt '*' kubeadm.join 10.160.65.165:6443 token='token'

    """
    cmd = ["kubeadm", "reset", "--force"]

    parameters = [
        ("cert-dir", cert_dir),
        ("cri-socket", cri_socket),
        ("ignore-preflight-errors", ignore_preflight_errors),
        ("kubeconfig", kubeconfig),
        ("rootfs", rootfs),
    ]
    for parameter, value in parameters:
        if value:
            cmd.extend([f"--{parameter}", str(value)])

    return _cmd(cmd)


# TODO(aplanas):
#   *  upgrade_apply
#   *  upgrade_diff
#   *  upgrade_node
#   *  upgrade_plan



"""
Test cases for salt.modules.win_pki
"""

import pytest

import salt.modules.win_pki as win_pki
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {win_pki: {}}


@pytest.fixture
def cert_path():
    return r"C:\certs\testdomain.local.cer"


@pytest.fixture
def thumbprint():
    return "9988776655443322111000AAABBBCCCDDDEEEFFF"


@pytest.fixture
def certs(thumbprint):
    return {
        thumbprint: {
            "dnsnames": ["testdomain.local"],
            "serialnumber": "0123456789AABBCCDD",
            "subject": "CN=testdomain.local, OU=testou, O=testorg, S=California, C=US",
            "thumbprint": thumbprint,
            "version": 3,
        }
    }


@pytest.fixture
def stores():
    return {
        "CurrentUser": [
            "AuthRoot",
            "CA",
            "ClientAuthIssuer",
            "Disallowed",
            "MSIEHistoryJournal",
            "My",
            "Root",
            "SmartCardRoot",
            "Trust",
            "TrustedPeople",
            "TrustedPublisher",
            "UserDS",
        ],
        "LocalMachine": [
            "AuthRoot",
            "CA",
            "ClientAuthIssuer",
            "Disallowed",
            "My",
            "Remote Desktop",
            "Root",
            "SmartCardRoot",
            "Trust",
            "TrustedDevices",
            "TrustedPeople",
            "TrustedPublisher",
            "WebHosting",
        ],
    }


@pytest.fixture
def json_certs():
    return [
        {
            "DnsNameList": [
                {"Punycode": "testdomain.local", "Unicode": "testdomain.local"}
            ],
            "SerialNumber": "0123456789AABBCCDD",
            "Subject": "CN=testdomain.local, OU=testou, O=testorg, S=California, C=US",
            "Thumbprint": "9988776655443322111000AAABBBCCCDDDEEEFFF",
            "Version": 3,
        }
    ]


@pytest.fixture
def json_stores(stores):
    return [
        {"LocationName": "CurrentUser", "StoreNames": stores["CurrentUser"]},
        {"LocationName": "LocalMachine", "StoreNames": stores["LocalMachine"]},
    ]


def test_get_stores(stores, json_stores):
    """
    Test - Get the certificate location contexts and their corresponding stores.
    """
    with patch.dict(win_pki.__salt__), patch(
        "salt.modules.win_pki._cmd_run", MagicMock(return_value=json_stores)
    ):
        assert win_pki.get_stores() == stores


def test_get_certs(certs, json_certs):
    """
    Test - Get the available certificates in the given store.
    """
    with patch.dict(win_pki.__salt__), patch(
        "salt.modules.win_pki._cmd_run", MagicMock(return_value=json_certs)
    ), patch("salt.modules.win_pki._validate_cert_path", MagicMock(return_value=None)):
        assert win_pki.get_certs() == certs


def test_get_cert_file(cert_path, thumbprint, certs, json_certs):
    """
    Test - Get the details of the certificate file.
    """
    kwargs = {"name": cert_path}
    with patch.dict(win_pki.__salt__), patch(
        "os.path.isfile", MagicMock(return_value=True)
    ), patch("salt.modules.win_pki._cmd_run", MagicMock(return_value=json_certs)):
        assert win_pki.get_cert_file(**kwargs) == certs[thumbprint]


def test_import_cert(cert_path, thumbprint, certs, json_certs):
    """
    Test - Import the certificate file into the given certificate store.
    """
    kwargs = {"name": cert_path}
    mock_value = MagicMock(return_value=cert_path)
    with patch.dict(win_pki.__salt__, {"cp.cache_file": mock_value}), patch(
        "salt.modules.win_pki._cmd_run", MagicMock(return_value=json_certs)
    ), patch(
        "salt.modules.win_pki._validate_cert_path", MagicMock(return_value=None)
    ), patch(
        "salt.modules.win_pki.get_cert_file",
        MagicMock(return_value=certs[thumbprint]),
    ), patch(
        "salt.modules.win_pki.get_certs", MagicMock(return_value=certs)
    ):
        assert win_pki.import_cert(**kwargs)


def test_export_cert(cert_path, thumbprint):
    """
    Test - Export the certificate to a file from the given certificate store.
    """
    kwargs = {"name": cert_path, "thumbprint": thumbprint}
    with patch.dict(win_pki.__salt__), patch(
        "salt.modules.win_pki._cmd_run", MagicMock(return_value="True")
    ), patch(
        "salt.modules.win_pki._validate_cert_format", MagicMock(return_value=None)
    ), patch(
        "salt.modules.win_pki._validate_cert_path", MagicMock(return_value=None)
    ):
        assert win_pki.export_cert(**kwargs)


def test_test_cert(thumbprint):
    """
    Test - Check the certificate for validity.
    """
    with patch.dict(win_pki.__salt__), patch(
        "salt.modules.win_pki._cmd_run", MagicMock(return_value="True")
    ), patch("salt.modules.win_pki._validate_cert_path", MagicMock(return_value=None)):
        assert win_pki.test_cert(thumbprint=thumbprint)


def test_remove_cert(thumbprint, certs):
    """
    Test - Remove the certificate from the given certificate store.
    """
    with patch.dict(win_pki.__salt__), patch(
        "salt.modules.win_pki._cmd_run", MagicMock(return_value=None)
    ), patch(
        "salt.modules.win_pki._validate_cert_path", MagicMock(return_value=None)
    ), patch(
        "salt.modules.win_pki.get_certs", MagicMock(return_value=certs)
    ):
        assert win_pki.remove_cert(thumbprint=thumbprint[::-1])


def test__cmd_run():
    """
    Test the _cmd_run function
    """
    mock_run = MagicMock(
        return_value={"retcode": 0, "stderr": "", "stdout": "some result"}
    )
    with patch.dict(win_pki.__salt__, {"cmd.run_all": mock_run}):
        result = win_pki._cmd_run(cmd="command")
        assert result == "some result"


def test__cmd_run_as_json():
    mock_run = MagicMock(
        return_value={"retcode": 0, "stderr": "", "stdout": '{"key": "value"}'}
    )
    with patch.dict(win_pki.__salt__, {"cmd.run_all": mock_run}):
        result = win_pki._cmd_run(cmd="command", as_json=True)
        assert result == {"key": "value"}


def test__cmd_run_stderr():
    mock_run = MagicMock(
        return_value={"retcode": 0, "stderr": "some error", "stdout": ""}
    )
    with patch.dict(win_pki.__salt__, {"cmd.run_all": mock_run}):
        with pytest.raises(CommandExecutionError) as exc_info:
            win_pki._cmd_run(cmd="command")
        expected = "Unable to execute command: command\nError: some error"
        assert exc_info.value.args[0] == expected


def test__cmd_run_bad_json():
    mock_run = MagicMock(
        return_value={"retcode": 0, "stderr": "", "stdout": "not : valid\njson"}
    )
    with patch.dict(win_pki.__salt__, {"cmd.run_all": mock_run}):
        with pytest.raises(CommandExecutionError) as exc_info:
            win_pki._cmd_run(cmd="command", as_json=True)
        expected = "Unable to parse return data as JSON:\nnot : valid\njson"
        assert exc_info.value.args[0] == expected



"""
Installing of mac pkg files
===========================

Install any kind of pkg, dmg or app file on macOS:

.. code-block:: yaml

    /mnt/test.pkg:
      macpackage.installed:
        - store: True

    /mnt/test.dmg:
      macpackage.installed:
        - dmg: True

    /mnt/xcode.dmg:
      macpackage.installed:
        - dmg: True
        - app: True
        - target: /Applications/Xcode.app
        - version_check: xcodebuild -version=Xcode 7.1\\n.*7B91b

"""

import logging
import os
import re

import salt.utils.platform
from salt.exceptions import CommandExecutionError

log = logging.getLogger(__name__)

__virtualname__ = "macpackage"


def __virtual__():
    """
    Only work on Mac OS
    """
    if salt.utils.platform.is_darwin():
        return __virtualname__
    return (False, "Only supported on Mac OS")


def installed(
    name,
    target="LocalSystem",
    dmg=False,
    store=False,
    app=False,
    mpkg=False,
    force=False,
    allow_untrusted=False,
    version_check=None,
):
    """
    Install a Mac OS Package from a pkg or dmg file, if given a dmg file it
    will first be mounted in a temporary location

    name
        The pkg or dmg file to install

    target
        The location in which to install the package. This can be a path or LocalSystem

    dmg
        Is the given file a dmg file?

    store
        Should the pkg be installed as if it was from the Mac OS Store?

    app
        Is the file a .app? If so then we'll just copy that to /Applications/ or the given
        target

    mpkg
        Is the file a .mpkg? If so then we'll check all of the .pkg files found are installed

    force
        Force the package to be installed even if its already been found installed

    allow_untrusted
        Allow the installation of untrusted packages

    version_check
        The command and version that we want to check against, the version number can use regex.

        .. code-block:: yaml

            version_check: python --version_check=2.7.[0-9]

    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}
    found = []
    installing = []

    real_pkg = name

    # Check version info
    if version_check is not None:
        split = version_check.split("=")
        if len(split) == 2:
            version_cmd = split[0]
            expected_version = split[1]
            try:
                version_out = __salt__["cmd.run"](
                    version_cmd, output_loglevel="quiet", ignore_retcode=True
                )
                version_out = version_out.strip()
            except CommandExecutionError:
                version_out = ""

            if re.match(expected_version, version_out) is not None:
                ret["comment"] += f"Version already matches {expected_version}"
                return ret
            else:
                ret["comment"] += "Version {} doesn't match {}. ".format(
                    version_out, expected_version
                )

    if app and target == "LocalSystem":
        target = "/Applications/"

    # Mount the dmg first
    mount_point = None
    if dmg:
        out, mount_point = __salt__["macpackage.mount"](name)
        if "attach failed" in out:
            ret["result"] = False
            ret["comment"] += f"Unable to mount {name}"
            return ret

        if app:
            real_pkg = mount_point + "/*.app"
        elif mpkg:
            real_pkg = mount_point + "/*.mpkg"
        else:
            real_pkg = mount_point + "/*.pkg"

    try:
        # Check if we have already installed this
        if app:
            if dmg:
                # Run with python shell due to the wildcard
                cmd = "ls -d *.app"
                out = __salt__["cmd.run"](cmd, cwd=mount_point, python_shell=True)

                if ".app" not in out:
                    ret["result"] = False
                    ret["comment"] += f"Unable to find .app in {mount_point}"
                    return ret
                else:
                    pkg_ids = out.split("\n")
            else:
                pkg_ids = [os.path.basename(name)]
                mount_point = os.path.dirname(name)

            if version_check is None:
                for p in pkg_ids:
                    if target[-4:] == ".app":
                        install_dir = target
                    else:
                        install_dir = os.path.join(target, p)
                    if os.path.exists(install_dir) and force is False:
                        found.append(p)
                    else:
                        installing.append(p)
            else:
                installing = pkg_ids
        else:
            installed_pkgs = __salt__["macpackage.installed_pkgs"]()

            if mpkg:
                pkg_ids = __salt__["macpackage.get_mpkg_ids"](real_pkg)
            else:
                pkg_ids = __salt__["macpackage.get_pkg_id"](real_pkg)

            if len(pkg_ids) > 0:
                for p in pkg_ids:
                    if p in installed_pkgs and force is False:
                        found.append(p)
                    else:
                        installing.append(p)
                if len(pkg_ids) == len(found):
                    return ret

        if app:

            def failed_pkg(f_pkg):
                ret["result"] = False
                ret["comment"] += f"{name} failed to install: {out}"

                if "failed" in ret["changes"]:
                    ret["changes"]["failed"].append(f_pkg)
                else:
                    ret["changes"]["failed"] = [f_pkg]

            for app in installing:
                try:
                    log.info("Copying %s to %s", app, target)

                    out = __salt__["macpackage.install_app"](
                        os.path.join(mount_point, app), target
                    )

                    if len(out) != 0:
                        failed_pkg(app)
                    else:
                        ret["comment"] += f"{app} installed"
                        if "installed" in ret["changes"]:
                            ret["changes"]["installed"].append(app)
                        else:
                            ret["changes"]["installed"] = [app]

                except OSError:
                    failed_pkg(app)
        else:
            out = __salt__["macpackage.install"](
                real_pkg, target, store, allow_untrusted
            )

            if out["retcode"] != 0:
                ret["result"] = False
                ret["comment"] += f". {name} failed to install: {out}"
            else:
                ret["comment"] += f"{name} installed"
                ret["changes"]["installed"] = installing

    finally:
        if dmg:
            # Unmount to be kind
            __salt__["macpackage.unmount"](mount_point)

    return ret



"""
The ssh client wrapper system contains the routines that are used to alter
how executions are run in the salt-ssh system, this allows for state routines
to be easily rewritten to execute in a way that makes them do the same tasks
as ZeroMQ salt, but via ssh.
"""

import copy

import salt.client.ssh
import salt.loader
import salt.utils.data
import salt.utils.json


class FunctionWrapper:
    """
    Create an object that acts like the salt function dict and makes function
    calls remotely via the SSH shell system
    """

    def __init__(
        self,
        opts,
        id_,
        host,
        wfuncs=None,
        mods=None,
        fsclient=None,
        cmd_prefix=None,
        aliases=None,
        minion_opts=None,
        **kwargs,
    ):
        super().__init__()
        self.cmd_prefix = cmd_prefix
        self.wfuncs = wfuncs if isinstance(wfuncs, dict) else {}
        self.opts = opts
        self.mods = mods if isinstance(mods, dict) else {}
        self.kwargs = {"id_": id_, "host": host}
        self.fsclient = fsclient
        self.kwargs.update(kwargs)
        self.aliases = aliases
        if self.aliases is None:
            self.aliases = {}
        self.minion_opts = minion_opts

    def __contains__(self, key):
        """
        We need to implement a __contains__ method, othwerwise when someone
        does a contains comparison python assumes this is a sequence, and does
        __getitem__ keys 0 and up until IndexError
        """
        try:
            self[key]  # pylint: disable=W0104
            return True
        except KeyError:
            return False

    def __getitem__(self, cmd):
        """
        Return the function call to simulate the salt local lookup system
        """
        if "." not in cmd and not self.cmd_prefix:
            # Form of salt.cmd.run in Jinja -- it's expecting a subdictionary
            # containing only 'cmd' module calls, in that case. Create a new
            # FunctionWrapper which contains the prefix 'cmd' (again, for the
            # salt.cmd.run example)
            kwargs = copy.deepcopy(self.kwargs)
            id_ = kwargs.pop("id_")
            host = kwargs.pop("host")
            return FunctionWrapper(
                self.opts,
                id_,
                host,
                wfuncs=self.wfuncs,
                mods=self.mods,
                fsclient=self.fsclient,
                cmd_prefix=cmd,
                aliases=self.aliases,
                minion_opts=self.minion_opts,
                **kwargs,
            )

        if self.cmd_prefix:
            # We're in an inner FunctionWrapper as created by the code block
            # above. Reconstruct the original cmd in the form 'cmd.run' and
            # then evaluate as normal
            cmd = f"{self.cmd_prefix}.{cmd}"

        if cmd in self.wfuncs:
            return self.wfuncs[cmd]

        if cmd in self.aliases:
            return self.aliases[cmd]

        def caller(*args, **kwargs):
            """
            The remote execution function
            """
            argv = [cmd]
            argv.extend([salt.utils.json.dumps(arg) for arg in args])
            argv.extend(
                [
                    "{}={}".format(
                        salt.utils.stringutils.to_str(key), salt.utils.json.dumps(val)
                    )
                    for key, val in kwargs.items()
                ]
            )
            single = salt.client.ssh.Single(
                self.opts,
                argv,
                mods=self.mods,
                disable_wipe=True,
                fsclient=self.fsclient,
                minion_opts=self.minion_opts,
                **self.kwargs,
            )
            stdout, stderr, retcode = single.cmd_block()
            if stderr.count("Permission Denied"):
                return {
                    "_error": "Permission Denied",
                    "stdout": stdout,
                    "stderr": stderr,
                    "retcode": retcode,
                }
            try:
                ret = salt.utils.json.loads(stdout)
                if len(ret) < 2 and "local" in ret:
                    ret = ret["local"]
                ret = ret.get("return", {})
            except ValueError:
                ret = {
                    "_error": "Failed to return clean data",
                    "stderr": stderr,
                    "stdout": stdout,
                    "retcode": retcode,
                }
            return ret

        return caller

    def __setitem__(self, cmd, value):
        """
        Set aliases for functions
        """
        if "." not in cmd and not self.cmd_prefix:
            # Form of salt.cmd.run in Jinja -- it's expecting a subdictionary
            # containing only 'cmd' module calls, in that case. We don't
            # support assigning directly to prefixes in this way
            raise KeyError(f"Cannot assign to module key {cmd} in the FunctionWrapper")

        if self.cmd_prefix:
            # We're in an inner FunctionWrapper as created by the first code
            # block in __getitem__. Reconstruct the original cmd in the form
            # 'cmd.run' and then evaluate as normal
            cmd = f"{self.cmd_prefix}.{cmd}"

        if cmd in self.wfuncs:
            self.wfuncs[cmd] = value

        # Here was assume `value` is a `caller` function from __getitem__.
        # We save it as an alias and then can return it when referenced
        # later in __getitem__
        self.aliases[cmd] = value

    def get(self, cmd, default):
        """
        Mirrors behavior of dict.get
        """
        if cmd in self:
            return self[cmd]
        else:
            return default



"""
    :codeauthor: Rahul Handay <rahulha@saltstack.com>
"""

import os

import pytest

import salt.states.svn as svn
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {svn: {}}


def test_latest():
    """
    Checkout or update the working directory to
    the latest revision from the remote repository.
    """
    mock = MagicMock(return_value=True)
    with patch.object(svn, "_fail", mock):
        assert svn.latest("salt")

    mock = MagicMock(side_effect=[True, False, False, False])
    with patch.object(os.path, "exists", mock):
        mock = MagicMock(return_value=False)
        with patch.object(os.path, "isdir", mock):
            with patch.object(svn, "_fail", mock):
                assert not svn.latest("salt", "c://salt")

        with patch.dict(svn.__opts__, {"test": True}):
            mock = MagicMock(return_value=["salt"])
            with patch.object(svn, "_neutral_test", mock):
                assert svn.latest("salt", "c://salt") == ["salt"]

            mock = MagicMock(side_effect=[False, True])
            with patch.object(os.path, "exists", mock):
                mock = MagicMock(return_value=True)
                info_mock = MagicMock(return_value=[{"Revision": "mocked"}])
                with patch.dict(
                    svn.__salt__, {"svn.diff": mock, "svn.info": info_mock}
                ):
                    mock = MagicMock(return_value=["Dude"])
                    with patch.object(svn, "_neutral_test", mock):
                        assert svn.latest("salt", "c://salt") == ["Dude"]

        with patch.dict(svn.__opts__, {"test": False}):
            mock = MagicMock(return_value=[{"Revision": "a"}])
            with patch.dict(svn.__salt__, {"svn.info": mock}):
                mock = MagicMock(return_value=True)
                with patch.dict(svn.__salt__, {"svn.update": mock}):
                    assert svn.latest("salt", "c://salt") == {
                        "changes": {},
                        "comment": True,
                        "name": "salt",
                        "result": True,
                    }


def test_latest_trust_failures():
    """
    Test that checks that the trust_failures option is handled
    correctly when running svn.latest in test mode. This tests for the
    bug reported as #59069.
    """
    os_path_exists_mock = MagicMock(side_effect=[False, True])
    svn_info_mock = MagicMock(return_value=[{"Revision": "42"}])
    svn_diff_mock = MagicMock()
    svn_neutral_test_mock = MagicMock()
    with patch.object(os.path, "exists", os_path_exists_mock), patch.dict(
        svn.__opts__, {"test": True}
    ), patch.dict(
        svn.__salt__, {"svn.diff": svn_diff_mock, "svn.info": svn_info_mock}
    ), patch.object(
        svn, "_neutral_test", svn_neutral_test_mock
    ):
        svn.latest("salt", "/my/test/dir", trust_failures="unknown-ca")
        svn_diff_mock.assert_called_with(
            "/my/test",
            "/my/test/dir",
            None,
            None,
            None,
            "-r",
            "42:HEAD",
            "--trust-server-cert-failures",
            "unknown-ca",
        )


def test_export():
    """
    Test to export a file or directory from an SVN repository
    """
    mock = MagicMock(return_value=True)
    with patch.object(svn, "_fail", mock):
        assert svn.export("salt")

    mock = MagicMock(side_effect=[True, False, False, False])
    with patch.object(os.path, "exists", mock):
        mock = MagicMock(return_value=False)
        with patch.object(os.path, "isdir", mock):
            with patch.object(svn, "_fail", mock):
                assert not svn.export("salt", "c://salt")

        with patch.dict(svn.__opts__, {"test": True}):
            mock = MagicMock(return_value=["salt"])
            with patch.object(svn, "_neutral_test", mock):
                assert svn.export("salt", "c://salt") == ["salt"]

            mock = MagicMock(side_effect=[False, True])
            with patch.object(os.path, "exists", mock):
                mock = MagicMock(return_value=True)
                with patch.dict(svn.__salt__, {"svn.list": mock}):
                    mock = MagicMock(return_value=["Dude"])
                    with patch.object(svn, "_neutral_test", mock):
                        assert svn.export("salt", "c://salt") == ["Dude"]

        with patch.dict(svn.__opts__, {"test": False}):
            mock = MagicMock(return_value=True)
            with patch.dict(svn.__salt__, {"svn.export": mock}):
                assert svn.export("salt", "c://salt") == {
                    "changes": {
                        "new": "salt",
                        "comment": "salt was Exported to c://salt",
                    },
                    "comment": True,
                    "name": "salt",
                    "result": True,
                }


def test_dirty():
    """
    Test to determine if the working directory has been changed.
    """
    mock = MagicMock(return_value=True)
    with patch.object(svn, "_fail", mock):
        assert svn.dirty("salt", "c://salt")



import pytest


@pytest.mark.slow_test
def test_config_items(salt_cli, salt_minion):
    ret = salt_cli.run("config.items", minion_tgt=salt_minion.id)
    assert ret.returncode == 0
    assert isinstance(ret.data, dict)



import pathlib

import pytest
from saltfactories.utils import random_string

import salt.modules.useradd as useradd
import salt.utils.platform
from salt.exceptions import CommandExecutionError

pytestmark = [
    pytest.mark.skip_if_not_root,
    pytest.mark.destructive_test,
    pytest.mark.windows_whitelisted,
]


@pytest.fixture(scope="module")
def user(modules):
    return modules.user


@pytest.fixture
def username(user):
    _username = random_string("test-account-", uppercase=False)
    try:
        yield _username
    finally:
        try:
            user.delete(_username, remove=True, force=True)
        except Exception:  # pylint: disable=broad-except
            # The point here is just system cleanup. It can fail if no account was created
            pass


@pytest.fixture
def account(username):
    with pytest.helpers.create_account(username=username) as account:
        yield account


def test_add(user, username):
    ret = user.add(username)
    assert ret is True


def test_delete(user, account):
    ret = user.delete(account.username)
    assert ret is True


@pytest.mark.skip_on_windows(reason="The windows user module does not support 'remove'")
@pytest.mark.parametrize("remove", [False, True])
def test_delete_remove(user, account, remove):
    """
    Test deleting a user from the system and passing ``remove`` to the call
    """
    user_info = user.info(account.username)
    ret = user.delete(account.username, remove=remove)
    assert ret is True
    if remove is True:
        assert pathlib.Path(user_info["home"]).exists() is False
    else:
        assert pathlib.Path(user_info["home"]).exists() is True


def test_info_after_deletion(user, account):
    """
    This test targets a situation where, at least on macOS, the call to ``user.info(username)``
    returns data after the account has been deleted from the system.
    It's a weird caching issue with ``pwd.getpwnam``
    """
    kwargs = {}
    if not salt.utils.platform.is_windows():
        kwargs["remove"] = True
    ret = user.delete(account.username, **kwargs)
    assert ret is True
    assert not user.info(account.username)


@pytest.mark.skip_unless_on_platforms(
    darwin=True,
    windows=True,
    reason="This test should only run (and raise an expected exception) on Windows/Mac due to lack of useradd command",
)
def test_errors_out_when_no_useradd_exists(username):
    """
    This test is fairly contrived, as we're calling the useradd module
    directly instead of letting the loader grab the appropriate user
    module for the given OS
    """
    with pytest.raises(CommandExecutionError):
        useradd.add(username)



"""
Runner functions supporting the Vault modules. Configuration instructions are
documented in the execution module docs.

:maintainer:    SaltStack
:maturity:      new
:platform:      all
"""

import base64
import copy
import json
import logging
import time
from collections.abc import Mapping

import requests

import salt.cache
import salt.crypt
import salt.exceptions
import salt.pillar
from salt.defaults import NOT_SET
from salt.exceptions import SaltRunnerError

log = logging.getLogger(__name__)


def generate_token(
    minion_id, signature, impersonated_by_master=False, ttl=None, uses=None
):
    """
    Generate a Vault token for minion minion_id

    minion_id
        The id of the minion that requests a token

    signature
        Cryptographic signature which validates that the request is indeed sent
        by the minion (or the master, see impersonated_by_master).

    impersonated_by_master
        If the master needs to create a token on behalf of the minion, this is
        True. This happens when the master generates minion pillars.

    ttl
        Ticket time to live in seconds, 1m minutes, or 2h hrs

    uses
        Number of times a token can be used
    """
    log.debug(
        "Token generation request for %s (impersonated by master: %s)",
        minion_id,
        impersonated_by_master,
    )
    _validate_signature(minion_id, signature, impersonated_by_master)
    try:
        config = __opts__.get("vault", {})
        verify = config.get("verify", None)
        # Vault Enterprise requires a namespace
        namespace = config.get("namespace")
        # Allow disabling of minion provided values via the master
        allow_minion_override = config["auth"].get("allow_minion_override", False)
        # This preserves the previous behavior of default TTL and 1 use
        if not allow_minion_override or uses is None:
            uses = config["auth"].get("uses", 1)
        if not allow_minion_override or ttl is None:
            ttl = config["auth"].get("ttl", None)
        storage_type = config["auth"].get("token_backend", "session")
        policies_refresh_pillar = config.get("policies_refresh_pillar", None)
        policies_cache_time = config.get("policies_cache_time", 60)

        if config["auth"]["method"] == "approle":
            if _selftoken_expired():
                log.debug("Vault token expired. Recreating one")
                # Requesting a short ttl token
                url = "{}/v1/auth/approle/login".format(config["url"])
                payload = {"role_id": config["auth"]["role_id"]}
                if "secret_id" in config["auth"]:
                    payload["secret_id"] = config["auth"]["secret_id"]
                # Vault Enterprise call requires headers
                headers = None
                if namespace is not None:
                    headers = {"X-Vault-Namespace": namespace}
                response = requests.post(
                    url, headers=headers, json=payload, verify=verify, timeout=120
                )
                if response.status_code != 200:
                    return {"error": response.reason}
                config["auth"]["token"] = response.json()["auth"]["client_token"]

        url = _get_token_create_url(config)
        headers = {"X-Vault-Token": config["auth"]["token"]}
        if namespace is not None:
            headers["X-Vault-Namespace"] = namespace
        audit_data = {
            "saltstack-jid": globals().get("__jid__", "<no jid set>"),
            "saltstack-minion": minion_id,
            "saltstack-user": globals().get("__user__", "<no user set>"),
        }
        payload = {
            "policies": _get_policies_cached(
                minion_id,
                config,
                refresh_pillar=policies_refresh_pillar,
                expire=policies_cache_time,
            ),
            "num_uses": uses,
            "meta": audit_data,
        }

        if ttl is not None:
            payload["explicit_max_ttl"] = str(ttl)

        if payload["policies"] == []:
            return {"error": "No policies matched minion"}

        log.trace("Sending token creation request to Vault")
        response = requests.post(
            url, headers=headers, json=payload, verify=verify, timeout=120
        )

        if response.status_code != 200:
            return {"error": response.reason}

        auth_data = response.json()["auth"]
        ret = {
            "token": auth_data["client_token"],
            "lease_duration": auth_data["lease_duration"],
            "renewable": auth_data["renewable"],
            "issued": int(round(time.time())),
            "url": config["url"],
            "verify": verify,
            "token_backend": storage_type,
            "namespace": namespace,
        }
        if uses >= 0:
            ret["uses"] = uses

        return ret
    except Exception as e:  # pylint: disable=broad-except
        return {"error": str(e)}


def unseal():
    """
    Unseal Vault server

    This function uses the 'keys' from the 'vault' configuration to unseal vault server

    vault:
      keys:
        - n63/TbrQuL3xaIW7ZZpuXj/tIfnK1/MbVxO4vT3wYD2A
        - S9OwCvMRhErEA4NVVELYBs6w/Me6+urgUr24xGK44Uy3
        - F1j4b7JKq850NS6Kboiy5laJ0xY8dWJvB3fcwA+SraYl
        - 1cYtvjKJNDVam9c7HNqJUfINk4PYyAXIpjkpN/sIuzPv
        - 3pPK5X6vGtwLhNOFv1U2elahECz3HpRUfNXJFYLw6lid

    .. note: This function will send unsealed keys until the api returns back
             that the vault has been unsealed

    CLI Examples:

    .. code-block:: bash

        salt-run vault.unseal
    """
    for key in __opts__["vault"]["keys"]:
        ret = __utils__["vault.make_request"](
            "PUT", "v1/sys/unseal", data=json.dumps({"key": key})
        ).json()
        if ret["sealed"] is False:
            return True
    return False


def show_policies(minion_id, refresh_pillar=NOT_SET, expire=None):
    """
    Show the Vault policies that are applied to tokens for the given minion.

    minion_id
        The minion's id.

    refresh_pillar
        Whether to refresh the pillar data when rendering templated policies.
        None will only refresh when the cached data is unavailable, boolean values
        force one behavior always.
        Defaults to config value ``policies_refresh_pillar`` or None.

    expire
        Policy computation can be heavy in case pillar data is used in templated policies and
        it has not been cached. Therefore, a short-lived cache specifically for rendered policies
        is used. This specifies the expiration timeout in seconds.
        Defaults to config value ``policies_cache_time`` or 60.

    CLI Example:

    .. code-block:: bash

        salt-run vault.show_policies myminion
    """
    config = __opts__.get("vault", {})
    if refresh_pillar == NOT_SET:
        refresh_pillar = config.get("policies_refresh_pillar")
    expire = expire if expire is not None else config.get("policies_cache_time", 60)
    return _get_policies_cached(
        minion_id, config, refresh_pillar=refresh_pillar, expire=expire
    )


def _validate_signature(minion_id, signature, impersonated_by_master):
    """
    Validate that either minion with id minion_id, or the master, signed the
    request
    """
    pki_dir = __opts__["pki_dir"]
    if impersonated_by_master:
        public_key = f"{pki_dir}/master.pub"
    else:
        public_key = f"{pki_dir}/minions/{minion_id}"

    log.trace("Validating signature for %s", minion_id)
    signature = base64.b64decode(signature)
    if not salt.crypt.verify_signature(public_key, minion_id, signature):
        raise salt.exceptions.AuthenticationError(
            f"Could not validate token request from {minion_id}"
        )
    log.trace("Signature ok")


# **kwargs because salt.cache.Cache does not pop "expire" from kwargs
def _get_policies(
    minion_id, config, refresh_pillar=None, **kwargs
):  # pylint: disable=unused-argument
    """
    Get the policies that should be applied to a token for minion_id
    """
    grains, pillar = _get_minion_data(minion_id, refresh_pillar)
    policy_patterns = config.get(
        "policies", ["saltstack/minion/{minion}", "saltstack/minions"]
    )
    mappings = {"minion": minion_id, "grains": grains, "pillar": pillar}

    policies = []
    for pattern in policy_patterns:
        try:
            for expanded_pattern in __utils__["vault.expand_pattern_lists"](
                pattern, **mappings
            ):
                policies.append(
                    expanded_pattern.format(**mappings).lower()  # Vault requirement
                )
        except KeyError:
            log.warning(
                "Could not resolve policy pattern %s for minion %s", pattern, minion_id
            )

    log.debug("%s policies: %s", minion_id, policies)
    return policies


def _get_policies_cached(minion_id, config, refresh_pillar=None, expire=60):
    # expiration of 0 disables cache
    if not expire:
        return _get_policies(minion_id, config, refresh_pillar=refresh_pillar)
    cbank = f"minions/{minion_id}/vault"
    ckey = "policies"
    cache = salt.cache.factory(__opts__)
    policies = cache.cache(
        cbank,
        ckey,
        _get_policies,
        expire=expire,
        minion_id=minion_id,
        config=config,
        refresh_pillar=refresh_pillar,
    )
    if not isinstance(policies, list):
        log.warning("Cached vault policies were not formed as a list. Refreshing.")
        cache.flush(cbank, ckey)
        policies = cache.cache(
            cbank,
            ckey,
            _get_policies,
            expire=expire,
            minion_id=minion_id,
            config=config,
            refresh_pillar=refresh_pillar,
        )
    return policies


def _get_minion_data(minion_id, refresh_pillar=None):
    _, grains, pillar = salt.utils.minions.get_minion_data(minion_id, __opts__)

    if grains is None:
        # In case no cached minion data is available, make sure the utils module
        # can distinguish a pillar refresh run impersonating a minion from running
        # on the master.
        grains = {"id": minion_id}
        # To properly refresh minion grains, something like this could be used:
        # __salt__["salt.execute"](minion_id, "saltutil.refresh_grains", refresh_pillar=False)
        # This is deliberately not done since grains should not be used to target
        # secrets anyways.

    # salt.utils.minions.get_minion_data only returns data from cache or None.
    # To make sure the correct policies are available, the pillar needs to be
    # refreshed. This can cause an infinite loop if the pillar data itself
    # depends on the vault execution module, which relies on this function.
    # By default, only refresh when necessary. Boolean values force one way.
    if refresh_pillar is True or (refresh_pillar is None and pillar is None):
        if __opts__.get("_vault_runner_is_compiling_pillar_templates"):
            raise SaltRunnerError(
                "Cyclic dependency detected while refreshing pillar for vault policy templating. "
                "This is caused by some pillar value relying on the vault execution module. "
                "Either remove the dependency from your pillar, disable refreshing pillar data "
                "for policy templating or do not use pillar values in policy templates."
            )
        local_opts = copy.deepcopy(__opts__)
        # Relying on opts for ext_pillars does not work properly (only the first one runs
        # correctly).
        extra_minion_data = {"_vault_runner_is_compiling_pillar_templates": True}
        local_opts.update(extra_minion_data)
        pillar = LazyPillar(
            local_opts, grains, minion_id, extra_minion_data=extra_minion_data
        )
    elif pillar is None:
        # Make sure pillar is a dict. Necessary because a check on LazyPillar would
        # refresh it unconditionally (even when no pillar values are used)
        pillar = {}

    return grains, pillar


def _selftoken_expired():
    """
    Validate the current token exists and is still valid
    """
    try:
        verify = __opts__["vault"].get("verify", None)
        # Vault Enterprise requires a namespace
        namespace = __opts__["vault"].get("namespace")
        url = "{}/v1/auth/token/lookup-self".format(__opts__["vault"]["url"])
        if "token" not in __opts__["vault"]["auth"]:
            return True
        headers = {"X-Vault-Token": __opts__["vault"]["auth"]["token"]}
        # Add Vault namespace to headers if Vault Enterprise enabled
        if namespace is not None:
            headers["X-Vault-Namespace"] = namespace
        response = requests.get(url, headers=headers, verify=verify, timeout=120)
        if response.status_code != 200:
            return True
        return False
    except Exception as e:  # pylint: disable=broad-except
        raise salt.exceptions.CommandExecutionError(
            f"Error while looking up self token : {str(e)}"
        )


def _get_token_create_url(config):
    """
    Create Vault url for token creation
    """
    role_name = config.get("role_name", None)
    auth_path = "/v1/auth/token/create"
    base_url = config["url"]
    return "/".join(x.strip("/") for x in (base_url, auth_path, role_name) if x)


class LazyPillar(Mapping):
    """
    Simulates a pillar dictionary. Only compiles the pillar
    once an item is requested.
    """

    def __init__(self, opts, grains, minion_id, extra_minion_data=None):
        self.opts = opts
        self.grains = grains
        self.minion_id = minion_id
        self.extra_minion_data = extra_minion_data or {}
        self._pillar = None

    def _load(self):
        log.info("Refreshing pillar for vault templating.")
        self._pillar = salt.pillar.get_pillar(
            self.opts,
            self.grains,
            self.minion_id,
            extra_minion_data=self.extra_minion_data,
        ).compile_pillar()

    def __getitem__(self, key):
        if self._pillar is None:
            self._load()
        return self._pillar[key]

    def __iter__(self):
        if self._pillar is None:
            self._load()
        yield from self._pillar

    def __len__(self):
        if self._pillar is None:
            self._load()
        return len(self._pillar)



import pytest

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.core_test,
]


def test_requisites_use(state, state_tree):
    """
    Call sls file containing several use_in and use.

    """
    # TODO issue #8235 & #8774 some examples are still commented in the test file
    sls_contents = """
    # None of theses states should run
    A:
      cmd.run:
        - name: echo "A"
        - onlyif: 'false'

    # issue #8235
    #B:
    #  cmd.run:
    #    - name: echo "B"
    #  # here used without "-"
    #    - use:
    #        cmd: A

    C:
      cmd.run:
        - name: echo "C"
        - use:
            - cmd: A

    D:
      cmd.run:
        - name: echo "D"
        - onlyif: 'false'
        - use_in:
            - cmd: E

    E:
      cmd.run:
        - name: echo "E"

    # issue 8235
    #F:
    #  cmd.run:
    #    - name: echo "F"
    #    - onlyif: return 0
    #    - use_in:
    #        cmd: G
    #
    #G:
    #  cmd.run:
    #    - name: echo "G"

    # issue xxxx
    #H:
    #  cmd.run:
    #    - name: echo "H"
    #    - use:
    #        - cmd: C
    #I:
    #  cmd.run:
    #    - name: echo "I"
    #    - use:
    #        - cmd: E
    """
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        for state_return in ret:
            assert state_return.comment == "onlyif condition is false"


@pytest.mark.skip(
    "issue #8802 : use recursions undetected. See comment in test for more info"
)
def test_requisites_use_recursion_1(state, state_tree):
    """
    Call sls file containing several use_in and use.
    """
    # TODO: issue #8802 : use recursions undetected
    # issue is closed as use does not actually inherit requisites
    # if chain-use is added after #8774 resolution theses tests would maybe become useful
    sls_contents = """
    A:
      cmd.run:
        - name: echo "A"
        - onlyif: return False
        - use:
            cmd: B

    B:
      cmd.run:
        - name: echo "B"
        - unless: return False
        - use:
            cmd: A
    """
    errmsg = 'A recursive requisite was found, SLS "requisite" ID "B" ID "A"'
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        assert ret.failed
        assert ret.errors == [errmsg]


@pytest.mark.skip(
    "issue #8802 : use recursions undetected. See comment in test for more info"
)
def test_requisites_use_recursion_2(state, state_tree):
    """
    Call sls file containing several use_in and use.
    """
    # TODO: issue #8802 : use recursions undetected
    # issue is closed as use does not actually inherit requisites
    # if chain-use is added after #8774 resolution theses tests would maybe become useful
    sls_contents = """
    #
    # A <--+ ---u--+
    #      |       |
    # B -u-+ <-+   |
    #          |   |
    # C -u-----+ <-+

    A:
      cmd.run:
        - name: echo "A"
        - use:
            cmd: C

    B:
      cmd.run:
        - name: echo "B"
        - use:
            cmd: C

    C:
      cmd.run:
        - name: echo "B"
        - use:
            cmd: A
    """
    errmsg = 'A recursive requisite was found, SLS "requisite" ID "C" ID "A"'
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        assert ret.failed
        assert ret.errors == [errmsg]


@pytest.mark.skip(
    "issue #8802 : use recursions undetected. See comment in test for more info"
)
def test_requisites_use_recursion_3(state, state_tree):
    """
    Call sls file containing several use_in and use.
    """
    # TODO: issue #8802 : use recursions undetected
    # issue is closed as use does not actually inherit requisites
    # if chain-use is added after #8774 resolution theses tests would maybe become useful
    sls_contents = """
    A:
      cmd.run:
        - name: echo "A"
        - use:
            cmd: A
    """
    errmsg = 'A recursive requisite was found, SLS "requisite" ID "A" ID "A"'
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        assert ret.failed
        assert ret.errors == [errmsg]


def test_requisites_use_no_state_module(state, state_tree):
    """
    Call sls file containing several use_in and use.
    """
    sls_contents = """
    # None of theses states should run
    A:
      cmd.run:
        - name: echo "A"
        - onlyif: 'false'

    # issue #8235
    #B:
    #  cmd.run:
    #    - name: echo "B"
    #  # here used without "-"
    #    - use:
    #        cmd: A

    C:
      cmd.run:
        - name: echo "C"
        - use:
            - A

    D:
      cmd.run:
        - name: echo "D"
        - onlyif: 'false'
        - use_in:
            - E

    E:
      cmd.run:
        - name: echo "E"

    # issue 8235
    #F:
    #  cmd.run:
    #    - name: echo "F"
    #    - onlyif: return 0
    #    - use_in:
    #        cmd: G
    #
    #G:
    #  cmd.run:
    #    - name: echo "G"

    # issue xxxx
    #H:
    #  cmd.run:
    #    - name: echo "H"
    #    - use:
    #        - cmd: C
    #I:
    #  cmd.run:
    #    - name: echo "I"
    #    - use:
    #        - cmd: E
    """
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        for state_return in ret:
            assert state_return.comment == "onlyif condition is false"



#!/usr/bin/env python

import sys


def main():
    print(
        "\nruntests.py support has been removed from Salt. Please try `nox -e"
        " 'pytest-3(coverage=True)'` "
        "or `nox -e 'pytest-3(coverage=True)' -- --help` to know more about the"
        " supported CLI flags.\n"
        "For more information, please check"
        " https://docs.saltproject.io/en/latest/topics/development/tests/index.html#running-the-tests",
        file=sys.stderr,
    )


if __name__ == "__main__":
    main()



"""
Integration tests for the Dimension Data cloud provider
"""

from tests.integration.cloud.helpers.cloud_test_base import TIMEOUT, CloudTest


class DimensionDataTest(CloudTest):
    """
    Integration tests for the Dimension Data cloud provider in Salt-Cloud
    """

    PROVIDER = "dimensiondata"
    REQUIRED_PROVIDER_CONFIG_ITEMS = ("key", "region", "user_id")

    def test_list_images(self):
        """
        Tests the return of running the --list-images command for the dimensiondata cloud provider
        """
        image_list = self.run_cloud(f"--list-images {self.PROVIDER}")
        self.assertIn("Ubuntu 14.04 2 CPU", [i.strip() for i in image_list])

    def test_list_locations(self):
        """
        Tests the return of running the --list-locations command for the dimensiondata cloud provider
        """
        _list_locations = self.run_cloud(f"--list-locations {self.PROVIDER}")
        self.assertIn(
            "Australia - Melbourne MCP2", [i.strip() for i in _list_locations]
        )

    def test_list_sizes(self):
        """
        Tests the return of running the --list-sizes command for the dimensiondata cloud provider
        """
        _list_sizes = self.run_cloud(f"--list-sizes {self.PROVIDER}")
        self.assertIn("default", [i.strip() for i in _list_sizes])

    def test_instance(self):
        """
        Test creating an instance on Dimension Data's cloud
        """
        # check if instance with salt installed returned
        ret_val = self.run_cloud(
            f"-p dimensiondata-test {self.instance_name}", timeout=TIMEOUT
        )
        self.assertInstanceExists(ret_val)

        self.assertDestroyInstance()



import salt.utils.master
from tests.support.mock import patch
from tests.support.unit import TestCase


class MasterPillarUtilTestCase(TestCase):
    """
    TestCase for salt.utils.master.MasterPillarUtil methods
    """

    def test_get_minion_pillar(self):
        """
        test get_minion_pillar when
        target exists
        """
        opts = {"test": False}
        minion = "minion"
        pillar = salt.utils.master.MasterPillarUtil(
            tgt=minion, tgt_type="glob", opts=opts
        )
        grains_data = {minion: {"domain": ""}}
        pillar_data = {minion: {"test_pillar": "foo"}}

        patch_grain = patch(
            "salt.utils.master.MasterPillarUtil._get_minion_grains",
            return_value=grains_data,
        )
        patch_pillar = patch(
            "salt.utils.master.MasterPillarUtil._get_minion_pillar",
            return_value=pillar_data,
        )
        patch_tgt_list = patch(
            "salt.utils.master.MasterPillarUtil._tgt_to_list", return_value=[minion]
        )

        with patch_grain, patch_pillar, patch_tgt_list:
            ret = pillar.get_minion_pillar()
        assert ret[minion] == pillar_data[minion]

    def test_get_minion_pillar_doesnotexist(self):
        """
        test get_minion_pillar when
        target does not exist
        """
        opts = {"test": False}
        minion = "minion"
        pillar = salt.utils.master.MasterPillarUtil(
            tgt="doesnotexist", tgt_type="glob", opts=opts
        )
        grains_data = {minion: {"domain": ""}}
        pillar_data = {minion: {"test_pillar": "foo"}}

        patch_grain = patch(
            "salt.utils.master.MasterPillarUtil._get_minion_grains",
            return_value=grains_data,
        )
        patch_pillar = patch(
            "salt.utils.master.MasterPillarUtil._get_minion_pillar",
            return_value=pillar_data,
        )
        patch_tgt_list = patch(
            "salt.utils.master.MasterPillarUtil._tgt_to_list", return_value=[]
        )

        with patch_grain, patch_pillar, patch_tgt_list:
            ret = pillar.get_minion_pillar()
        assert minion not in ret

    def test_get_minion_pillar_notgt(self):
        """
        test get_minion_pillar when
        passing target None
        """
        opts = {"test": False}
        minion = "minion"
        pillar = salt.utils.master.MasterPillarUtil(
            tgt=None, tgt_type="glob", opts=opts
        )
        grains_data = {minion: {"domain": ""}}
        pillar_data = {minion: {"test_pillar": "foo"}}

        patch_grain = patch(
            "salt.utils.master.MasterPillarUtil._get_minion_grains",
            return_value=grains_data,
        )
        patch_pillar = patch(
            "salt.utils.master.MasterPillarUtil._get_minion_pillar",
            return_value=pillar_data,
        )
        patch_tgt_list = patch(
            "salt.utils.master.MasterPillarUtil._tgt_to_list", return_value=[]
        )

        with patch_grain, patch_pillar, patch_tgt_list:
            ret = pillar.get_minion_pillar()
        assert minion in ret



"""
Specifically ensure that pillars are merged as expected
for the target as well and available for renderers.
This should be covered by `test.check_pillar` above, but
let's check the specific output for the most important funcs.
Issue #59802
"""

import json

import pytest

import salt.utils.dictupdate

pytestmark = [
    pytest.mark.skip_on_windows(reason="salt-ssh not available on Windows"),
    pytest.mark.usefixtures("pillar_tree_nested"),
    pytest.mark.slow_test,
]


@pytest.fixture
def _write_pillar_state(base_env_state_tree_root_dir, tmp_path_factory):
    tmp_path = tmp_path_factory.mktemp("tgtdir")
    tgt_file = tmp_path / "deepthought.txt"
    top_file = """
    base:
      'localhost':
        - writepillar
      '127.0.0.1':
        - writepillar
    """
    # Inline the Jinja template content to avoid salt-ssh file server lookup issues
    # while still testing Jinja rendering with pillar data
    deepthought_template = r"""{{
      {
        "raw": {
          "the_meaning": pillar.get("the_meaning"),
          "btw": pillar.get("btw")},
        "modules": {
          "the_meaning": salt["pillar.get"]("the_meaning"),
          "btw": salt["pillar.get"]("btw")}
      } | json }}"""
    nested_pillar_file = f"""
    deep_thought:
      file.managed:
        - name: {tgt_file}
        - contents: |
            {deepthought_template}
        - template: jinja
    """
    top_tempfile = pytest.helpers.temp_file(
        "top.sls", top_file, base_env_state_tree_root_dir
    )
    show_tempfile = pytest.helpers.temp_file(
        "writepillar.sls", nested_pillar_file, base_env_state_tree_root_dir
    )

    with top_tempfile, show_tempfile:
        yield tgt_file


@pytest.fixture
def base():
    return {"the_meaning": {"of": {"life": 42, "bar": "tender"}, "for": "what"}}


@pytest.fixture
def override(base):
    poverride = {
        "the_meaning": {"of": {"life": 2.71, "foo": "lish"}},
        "btw": "turtles",
    }
    expected = salt.utils.dictupdate.merge(base, poverride)
    return expected, poverride


@pytest.mark.parametrize(
    "args,kwargs",
    (
        (("state.sls", "writepillar"), {}),
        (("state.highstate",), {"whitelist": "writepillar"}),
        (("state.top", "top.sls"), {}),
    ),
)
def test_it(salt_ssh_cli, args, kwargs, override, _write_pillar_state):
    expected, override = override
    ret = salt_ssh_cli.run(*args, **kwargs, pillar=override)
    assert ret.returncode == 0
    assert isinstance(ret.data, dict)
    assert ret.data
    assert _write_pillar_state.exists()
    pillar = json.loads(_write_pillar_state.read_text())
    assert pillar["raw"] == expected
    assert pillar["modules"] == expected



import copy
import datetime
import logging

log = logging.getLogger(__name__)


def test_get_schedule_remove_hidden_true(schedule):
    """
    verify that the _get_schedule function works
    when remove_hidden is True and schedule data
    contains enabled key and that opts["schedule"]
    is not changed.
    """

    job_name = "test_get_schedule"
    job = {
        "schedule": {
            "enabled": True,
            job_name: {
                "function": "test.ping",
                "seconds": 60,
                "_next_fire_time": datetime.datetime(2023, 2, 13, 18, 25, 16, 271796),
                "_splay": None,
                "_seconds": 3600,
                "_next_scheduled_fire_time": datetime.datetime(
                    2023, 2, 13, 18, 25, 16, 271796
                ),
                "_skip_reason": "disabled",
                "_skipped_time": datetime.datetime(2023, 2, 13, 17, 26, 16, 271381),
                "_skipped": True,
            },
        }
    }

    expected_ret = {
        "enabled": True,
        job_name: {"function": "test.ping", "seconds": 60},
    }
    # Add the job to the scheduler
    schedule.opts.update(copy.deepcopy(job))

    ret = schedule._get_schedule(remove_hidden=True)
    assert expected_ret == ret
    assert schedule.opts["schedule"][job_name] == job["schedule"][job_name]


def test_get_schedule_remove_hidden_false(schedule):
    """
    verify that the _get_schedule function works
    when remove_hidden is False and schedule data
    contains enabled key and that opts["schedule"]
    is not changed.
    """

    job_name = "test_get_schedule"
    job = {
        "schedule": {
            "enabled": True,
            job_name: {
                "function": "test.ping",
                "seconds": 60,
                "_next_fire_time": datetime.datetime(2023, 2, 13, 18, 25, 16),
                "_splay": None,
                "_seconds": 3600,
                "_next_scheduled_fire_time": datetime.datetime(2023, 2, 13, 18, 25, 16),
                "_skip_reason": "disabled",
                "_skipped_time": datetime.datetime(2023, 2, 13, 17, 26, 16),
                "_skipped": True,
            },
        }
    }
    # Add the job to the scheduler
    schedule.opts.update(copy.deepcopy(job))

    ret = schedule._get_schedule(remove_hidden=False)
    assert job["schedule"] == ret



import platform

import pytest


@pytest.fixture(scope="module")
def account():
    with pytest.helpers.create_account(create_group=True) as _account:
        yield _account


@pytest.mark.skip_on_windows
@pytest.mark.destructive_test
@pytest.mark.skip_if_not_root
def test_runas_id(cmd, account):
    ret = cmd.run("id", runas=account.username)
    assert ret.result is True
    assert f"uid={account.info.uid}" in ret.changes["stdout"]
    assert f"gid={account.info.gid}" in ret.changes["stdout"]


@pytest.mark.windows_whitelisted
@pytest.mark.skip_unless_on_windows
def test_runas_whoami(cmd, account):
    ret = cmd.run("whoami /user /groups /fo list", runas=account.username)
    hostname = platform.node().lower()
    assert (
        f"User Name: {hostname}\\{account.username}".lower()
        in ret.changes["stdout"].lower()
    )
    assert (
        f"Group Name: {hostname}\\{account.group_name}".lower()
        in ret.changes["stdout"].lower()
    )



"""
Tests for file.symlink function
"""

import os

import pytest

import salt.utils.path
from salt.exceptions import CommandExecutionError, SaltInvocationError

pytestmark = [
    pytest.mark.windows_whitelisted,
]


@pytest.fixture(scope="module")
def file(modules):
    return modules.file


@pytest.fixture(scope="function")
def source():
    with pytest.helpers.temp_file(contents="Source content") as source:
        yield source


def test_symlink(file, source):
    """
    Test symlink with defaults
    """
    target = source.parent / "symlink.lnk"
    try:
        file.symlink(str(source), str(target))
        assert salt.utils.path.islink(str(target))
    finally:
        target.unlink()


def test_symlink_missing_src(file, source):
    """
    Test symlink when src is missing should still create the link
    """
    target = source.parent / "symlink.lnk"
    missing_source = source.parent / "missing.txt"
    try:
        file.symlink(str(missing_source), str(target))
        assert salt.utils.path.islink(str(target))
    finally:
        target.unlink()


def test_symlink_exists_same(file, source):
    """
    Test symlink with an existing symlink to the correct file
    Timestamps should not change
    """
    target = source.parent / "symlink.lnk"
    target.symlink_to(source)
    try:
        before_time = os.stat(str(target)).st_mtime
        ret = file.symlink(str(source), str(target))
        after_time = os.stat(str(target)).st_mtime
        assert before_time == after_time
        assert ret is True
    finally:
        target.unlink()


def test_symlink_exists_different(file, source):
    """
    Test symlink with an existing symlink to a different file
    Should throw a CommandExecutionError
    """
    dif_source = source.parent / "dif_source.txt"
    target = source.parent / "symlink.lnk"
    target.symlink_to(dif_source)
    try:
        with pytest.raises(CommandExecutionError) as exc:
            file.symlink(str(source), str(target))
        assert "Found existing symlink:" in exc.value.message
    finally:
        target.unlink()


def test_symlink_exists_file(file, source):
    """
    Test symlink when the existing file is not a link
    We don't do anything because we do not want to destroy any data
    Should throw a CommandExecutionError
    """
    with pytest.helpers.temp_file("symlink.txt", contents="Source content") as target:
        with pytest.raises(CommandExecutionError) as exc:
            file.symlink(str(source), str(target))
        assert "Existing path is not a symlink:" in exc.value.message


def test_symlink_exists_different_force(file, source):
    """
    Test symlink with an existing symlink to a different file with force=True
    Should destroy the existing symlink and generate a new one to the correct
    location
    """
    dif_source = source.parent / "dif_source.txt"
    target = source.parent / "symlink.lnk"
    target.symlink_to(dif_source)
    try:
        file.symlink(str(source), str(target), force=True)
        assert salt.utils.path.readlink(str(target)) == str(source)
    finally:
        target.unlink()


def test_symlink_target_relative_path(file, source):
    """
    Test symlink when the target file is a relative path
    Should throw a SaltInvocationError
    """
    target = f"..{os.path.sep}symlink.lnk"
    with pytest.raises(SaltInvocationError) as exc:
        file.symlink(str(source), str(target))
    assert "Link path must be absolute" in exc.value.message


def test_symlink_exists_different_atomic(file, source):
    """
    Test symlink with an existing symlink to a different file with atomic=True
    Should replace the existing symlink with a new one to the correct location
    """
    dif_source = source.parent / "dif_source.txt"
    target = source.parent / "symlink.lnk"
    target.symlink_to(dif_source)
    try:
        file.symlink(str(source), str(target), atomic=True)
        assert salt.utils.path.readlink(str(target)) == str(source)
    finally:
        target.unlink()



import multiprocessing
import time

import salt.ext.tornado.gen
import salt.transport.tcp


async def test_pub_channel(master_opts, minion_opts, io_loop):
    def presence_callback(client):
        pass

    def remove_presence_callback(client):
        pass

    master_opts["transport"] = "tcp"
    minion_opts.update(master_ip="127.0.0.1", transport="tcp")

    server = salt.transport.tcp.TCPPublishServer(master_opts)

    client = salt.transport.tcp.TCPPubClient(minion_opts, io_loop)

    payloads = []

    publishes = []

    def publish_payload(payload, callback):
        server.publish_payload(payload)
        payloads.append(payload)

    def on_recv(message):
        publishes.append(message)

    proc = multiprocessing.Process(
        target=server.publish_daemon,
        args=(publish_payload, presence_callback, remove_presence_callback),
    )
    proc.start()

    # Wait for socket to bind.
    time.sleep(3)

    await client.connect(master_opts["publish_port"])
    client.on_recv(on_recv)

    server.publish({"meh": "bah"})

    start = time.monotonic()
    try:
        while not publishes:
            await salt.ext.tornado.gen.sleep(0.3)
            if time.monotonic() - start > 30:
                assert False, "Message not published after 30 seconds"
    finally:
        client.close()
        server.close()
        proc.terminate()
        proc.join()



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.states.ddns as ddns
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {ddns: {}}


def test_present():
    """
    Test to ensures that the named DNS record is present with the given ttl.
    """
    name = "webserver"
    zone = "example.com"
    ttl = "60"
    data = "111.222.333.444"

    ret = {"name": name, "result": None, "comment": "", "changes": {}}

    with patch.dict(ddns.__opts__, {"test": True}):
        comt = f'A record "{name}" will be updated'
        ret.update({"comment": comt})
        assert ddns.present(name, zone, ttl, data) == ret

        with patch.dict(ddns.__opts__, {"test": False}):
            mock = MagicMock(return_value=None)
            with patch.dict(ddns.__salt__, {"ddns.update": mock}):
                comt = f'A record "{name}" already present with ttl of {ttl}'
                ret.update({"comment": comt, "result": True})
                assert ddns.present(name, zone, ttl, data) == ret


def test_absent():
    """
    Test to ensures that the named DNS record is absent.
    """
    name = "webserver"
    zone = "example.com"
    data = "111.222.333.444"

    ret = {"name": name, "result": None, "comment": "", "changes": {}}

    with patch.dict(ddns.__opts__, {"test": True}):
        comt = f'None record "{name}" will be deleted'
        ret.update({"comment": comt})
        assert ddns.absent(name, zone, data) == ret

        with patch.dict(ddns.__opts__, {"test": False}):
            mock = MagicMock(return_value=None)
            with patch.dict(ddns.__salt__, {"ddns.delete": mock}):
                comt = "No matching DNS record(s) present"
                ret.update({"comment": comt, "result": True})
                assert ddns.absent(name, zone, data) == ret
